# -*- coding: utf-8 -*-
"""
keri.core.indexing module

Provides versioning support for Indexer classes and codes
"""


from typing import NamedTuple
from collections import namedtuple
from collections.abc import Mapping
from dataclasses import dataclass, astuple, asdict

from ..kering import (EmptyMaterialError, InvalidValueError,)

from .. import help
from ..help import nonStringSequence

from . import coring
from .coring import (MapDom, Matter, Diger, Prefixer, Number)



# ToDo Change seal namedtuple definitions to NamedTuple subclasses so can
# use typehints on field values which type hints are the primitive types. Use
# union | on type hints to allow qb64, qb2, primitive instance, primitive class
# as acceptable values  This provides more clarity in documentation. Actually
# enforcing types is harder with union | but can still be accomplished.

#  for the following Seal namedtuples use the ._asdict() method to convert to dict
#  when using in events
# to convert seal namedtuple to dict use namedtuple._asdict()
# seal == SealEvent(i="abc",s="1",d="efg")
# sealdict =seal._asdict()
# to convet dict to namedtuple use ** unpacking as in seal = SealDigest(**sealdict)
# to check if dict of seal matches fields of associted namedtuple
# if tuple(sealdict) == SealEvent._fields:

# Digest Seal: uniple (d,)
# d = digest qb64 of data  (usually SAID)
SealDigest = namedtuple("SealDigest", 'd')

# Root Seal: uniple (rd,)
# rd = Merkle tree root digest qb64 digest of anchored (sealed) data in Merkle tree
SealRoot = namedtuple("SealRoot", 'rd')

# Backer Seal: couple (bi, d)
# bi = pre qb64 backer nontrans identifier prefix
# d = digest qb64 of backer metadata anchored to event usually SAID of data
SealBacker = namedtuple("SealBacker", 'bi d')

# Last Establishment Event Seal: uniple (i,)
# i = pre is qb64 of identifier prefix of KEL from which to get last est, event
# used to indicate to get the latest keys available from KEL for 'i'
SealLast = namedtuple("SealLast", 'i')

# Transaction Event Seal for Transaction Event: duple (s, d)
# s = sn of transaction event as lowercase hex string  no leading zeros,
# d = SAID digest qb64 of transaction event
# the pre is provided in the 'i' field  qb64 of identifier prefix of KEL
# key event that this seal appears.
# use SealSourceCouples count code for attachment
SealTrans = namedtuple("SealTrans", 's d')

# Event Seal: triple (i, s, d)
# i = pre is qb64 of identifier prefix of KEL for event,
# s = sn of event as lowercase hex string  no leading zeros,
# d = SAID digest qb64 of event
SealEvent = namedtuple("SealEvent", 'i s d')


# Following are not seals only used in database

# State Establishment Event (latest current) : quadruple (s, d, br, ba)
# s = sn of latest est event as lowercase hex string  no leading zeros,
# d = SAID digest qb64  of latest establishment event
# br = backer (witness) remove list (cuts) from latest est event
# ba = backer (witness) add list (adds) from latest est event
StateEstEvent = namedtuple("StateEstEvent", 's d br ba')

# not used should this be depricated?
# State Event (latest current) : triple (s, t, d)
# s = sn of latest event as lowercase hex string  no leading zeros,
# t = message type of latest event (ilk)
# d = SAID digest qb64 of latest event
StateEvent = namedtuple("StateEvent", 's t d')


# Cast conversion: duple (c, k)
# kls = primitive class reference in order to cast as appropriate
#       namedtuple with values as primitive classes
# prm = primitive __init__ keyword parameter name to use when casting
#        default None. When default then use qb64 or qb64b as appropriate.
Castage = namedtuple('Castage', "kls prm", defaults=(None, ))


@dataclass(frozen=True)
class EmptyClanDom(MapDom):
    """
    SealClanDom is dataclass of namedtuple seal class references (clans) each
    indexed by its class name.

    Only provide defined classes.
    Undefined are left out so that inclusion(exclusion) via 'in' operator works.

    As subclass of MapCodex can get class reference with item syntax using
    name variables.

    Example: EmptyClanDex[name]
    """

    def __iter__(self):
        return iter(astuple(self))  # enables value not key inclusion test with "in"

EClanDom = EmptyClanDom()  # create instance


@dataclass(frozen=True)
class EmptyCastDom(MapDom):
    """
    SealCastCodex is dataclass of namedtuple instances (seal casts) whose values
    are named primitive class references

    indexed by its namedtuple class name.

    Only provide defined namedtuples casts.
    Undefined are left out so that inclusion(exclusion) via 'in' operator works.

    As subclass of MapCodex can get namedtuple instance with item syntax using
    name variables.

    Example: EmptyCastDex[name]
    """

    def __iter__(self):
        return iter(astuple(self))  # enables value not key inclusion test with "in"

ECastDom = EmptyCastDom()  # create instance


@dataclass(frozen=True)
class SealClanDom(MapDom):
    """
    SealClanDom is dataclass of namedtuple seal class references (clans) each
    indexed by its class name.

    Only provide defined classes.
    Undefined are left out so that inclusion(exclusion) via 'in' operator works.

    As subclass of MapCodex can get class reference with item syntax using
    name variables.

    Example: ClanDom[name]
    """
    SealDigest: type[NamedTuple] = SealDigest  # SealDigest class reference
    SealRoot: type[NamedTuple] = SealRoot  # SealRoot class reference
    SealBacker: type[NamedTuple] = SealBacker  # SealBacker class reference
    SealLast: type[NamedTuple] = SealLast  # SealLast class reference single
    SealTrans: type[NamedTuple] = SealTrans  # SealTrans class reference couple
    SealEvent: type[NamedTuple] = SealEvent  # SealEvent class reference triple

    def __iter__(self):
        return iter(astuple(self))  # enables value not key inclusion test with "in"

SClanDom = SealClanDom()  # create instance




@dataclass(frozen=True)
class SealCastDom(MapDom):
    """
    SealCastDom is dataclass of namedtuple instances (seal casts) whose values
    are named primitive class references

    indexed by its namedtuple class name.

    Only provide defined namedtuples casts.
    Undefined are left out so that inclusion(exclusion) via 'in' operator works.

    As subclass of MapCodex can get namedtuple instance with item syntax using
    name variables.

    Example: CastDom[name]
    """
    SealDigest: NamedTuple = SealDigest(d=Castage(Diger))  # SealDigest class reference
    SealRoot: NamedTuple = SealRoot(rd=Castage(Diger))  # SealRoot class reference
    SealBacker: NamedTuple = SealBacker(bi=Castage(Prefixer),
                                        d=Castage(Diger))  # SealBacker class reference
    SealLast: NamedTuple = SealLast(i=Castage(Prefixer))  # SealLast class reference single
    SealTrans: NamedTuple = SealTrans(s=Castage(Number, 'numh'),
                                      d=Castage(Diger))  # SealTrans class reference couple
    SealEvent: NamedTuple = SealEvent(i=Castage(Prefixer),
                                      s=Castage(Number, 'numh'),
                                      d=Castage(Diger))  # SealEvent class reference triple

    def __iter__(self):
        return iter(astuple(self))  # enables value not key inclusion test with "in"

SCastDom = SealCastDom()  # create instance



class Structor:
    """Structor class each instance holds a namedtuple .data of named values.
    Each value is a primitive instance of CESR primitive subclass that supports
    text (ab64) and binary (qb2) domains.
    Structor instances can be serialized to or deserialized from concatenation
    of the qb64 or qb2 representations of the data values. Creation requires
    input of a ordered named classes for creating the named instances from the
    input data. Smart data format input are supported to accomodate the many ways
    the named data may appear in messages and or databases.

    Instance Creation Patterns:

        Structor(data):

        Structor(clan, cast, crew):
        Structor(clan, cast, qb64):
        Structor(clan, cast, qb2):

        Structor(cast, crew):
        Structor(cast, qb64):
        Structor(cast, qb2):

        Structor(clan, crew):  when known cast in .Casts for clan
        Structor(clan, qb64): when known cast in .Casts for clan
        Structor(clan, qb2):  when known cast in .Casts for clan

        Structor(crew): when known cast in .Casts for crew


    Class Attributes:
        Clans (type[Namedtuple]): each value is known NamedTuple class keyed
            by its own field names (tuple). Enables easy query of its values() to
            find known data types given field names tuple.

        Casts (NamedTuple): each value is primitive class of cast keyed by fields
            names of the associated NamedTuple class in .Clans. Enables finding
            known primitive classes given NamedTuple class of clan or instance
            of cast or crew.

    When known casts or provided in .Clans/.Casts then more flexible creation
    is supported for different types of provided cast and crew.
    When no clan is provided and an unknown cast and/or crew are provided as
    Mappings then Structor may create custom clan from the names given by the
    cast and/or crew keys(). Subclasses may override this behavior by raising
    an exception for unknown or custom clans.


    Properties:
        data (NamedTuple): fields are named instances of CESR primitives
        clan (type[NamedTuple]): class reference of .data's class
        cast (NamedTuple | None): values are Castage instances that each provide
                    CESR primitive class references and primitive init parameter
                    used to .data's primitive instances.
        crew (NamedTuple): named qb64 values of .data's primitive instances
        qb64 (str): concatenated data values as qb64 str of data's primitives
        qb64b (bytes): concatenated data values as qb64b  of data's primitives
        qb2 (bytes): concatenated data values as qb2 bytes of data's primitives


    Methods:


    Hidden:
        _data (NamedTuple): named CESR primitive instances


    """
    Clans = EClanDom  # known namedtuple clans. Override in subclass with non-empty
    Casts = ECastDom  # known namedtuple casts. Override in subclass with non-empty
    # Create .Names dict that maps tuple of clan/cast fields names to its namedtuple
    # class type name so can look up a know clan or cast given a matching tuple
    # of either field names from a namedtuple or keys from a dict. The tuple of
    # field names is a mark of the structor type. Maps mark to class name.
    Names = {tuple(clan._fields): clan.__name__ for clan in Clans}


    def __init__(self, data=None, *, clan=None, cast=None, crew=None,
                 qb64=None, qb2=None, strip=False):
        """Initialize instance

        Parameters:
            data (NamedTuple | None): fields are named primitive instances for .data
                Given data can derive clan, cast, crew, qb64, and qb2
            clan (type[NamedTuple]): provides class reference for generated .data
                when data missing.
            cast (NamedTuple | dict | Iterable | None):  values are Castage
                    instances that each provide CESR primitive class references
                    and primitive init parameter used to .data's primitive
                    instances. None means .data provided directly not generated
                    from cast
            each value provides CESR
                primitive subclass reference used to create primitive instances
                for generating .data. Can be used to infer namedtuple type of
                .data when data and clan missing. Takes precendence over crew.
            crew (NamedTuple | dict | Iterable | None): each value provides qb64 value
                of primitive for generating .data with .cast when data missing.
                Can be used to infer namedtuple type of .data when data and clan
                missing.
            qb64 (str | bytes | bytearray | None): concatenation of qb64 data values to
                generate .data with data and crew missing.
            qb2 (bytes | bytearray | None): concatenation of qb2 data values to generate
                .data when data and crew and qb64 missing.
            strip (bool): False means do not strip each value from qb64 or qb2.
                            Default is False.
                          True means if qb64 or qb2 are bytearray then strip
                            contained concatenated data values. Enables parser
                            to extract data fields from front of CESR stream.


        """
        if data:
            if not (isinstance(data, tuple) and hasattr(data, "_fields")):
                raise InvalidValueError(f"Not namedtuple subclass {data=}.")

            for pi in data:  # check for primitive interface
                if not (hasattr(pi, "qb64") and hasattr(pi, "qb2")):
                    raise InvalidValueError(f"Non-primitive data member={pi}.")

            cast = None  # ensure cast is None since not used to generate data


        else:
            if not clan:  # attempt to get from cast and/or crew
                if cast and isinstance(cast, tuple) and hasattr(cast, "_fields"):
                    clan = cast.__class__

                if not clan and crew:
                    if isinstance(crew, tuple) and hasattr(crew, "_fields"):
                        clan = crew.__class__

                if not clan and isinstance(cast, Mapping):  # get clan from cast
                    mark = tuple(cast)  # create custom clan based on cast mark
                    if (cname := self.Names.get(mark)):  # get known else None
                        clan = self.Clans[cname]
                        cast = self.Casts[cname]

                    else:  # create custom clan from cast
                        clan = namedtuple("_".join(mark), mark)  # custom clan

                if not clan and isinstance(crew, Mapping):  # get clan from crew
                    mark = tuple(crew)  # create custom clan based on crew mark
                    if (cname := self.Names.get(mark)):  # get known else None
                        clan = self.Clans[cname]
                        cast = self.Casts[cname]

                    else:  # create custom clan from crew
                        clan = namedtuple("_".join(mark), mark)  # custom clan from cast keys

            if clan:
                if not (issubclass(clan, tuple) and hasattr(clan, "_fields")):
                    raise InvalidValueError(f"Not namedtuple subclass {clan=}.")
            else:
                raise InvalidValueError(f"Missing or unobtainable clan.")

            # have clan but may not have cast
            if cast:
                if not isinstance(cast, clan):
                    if isinstance(cast, tuple) and hasattr(cast, "_fields"):
                        if cast._fields != clan._fields:  # fields is mark
                            raise InvalidValueError(f"Mismatching fields clan="
                                                    f"{clan._fields} and cast="
                                                    f"{cast._fields}.")

                        cast = clan(**cast._asdict())  # convert to clan

                    elif isinstance(cast, Mapping):
                        if tuple(cast) != clan._fields:  # fields is mark
                            raise InvalidValueError(f"Mismatching fields clan="
                                                    f"{clan._fields} and keys cast="
                                                    f"{tuple(cast)}.")

                        cast = clan(**cast)  # convert to clan

                    elif isinstance(cast, nonStringSequence):
                        cast = clan(*cast)  # convert to clan assumes elements in correct order

                    else:
                        raise InvalidValueError(f"Invalid {cast=}.")

            else:  # get cast from known .Casts if possible
                if (cname := self.Names.get(clan._fields)):  # fields is mark
                    cast = self.Casts[cname]  # get known cast
                else:  # cast missing or unobtainable
                    raise InvalidValueError(f"Missing or unobtainable cast.")

            # have cast now
            for cstg in cast:
                if not (hasattr(cstg.kls, "qb64") and hasattr(cstg.kls, "qb2")):
                    raise InvalidValueError(f"Cast member {cstg.kls=} not CESR"
                                            " Primitive.")

            # have clan and cast but may not have crew
            if crew:
                if not isinstance(crew, clan):
                    if isinstance(crew, tuple) and hasattr(crew, "_fields"):
                        if crew._fields != clan._fields:  # fields is mark
                            raise InvalidValueError(f"Mismatching fields clan="
                                                    f"{clan._fields} and crew="
                                                    f"{crew._fields}.")

                        crew = clan(**crew._asdict())  # convert to clan

                    elif isinstance(crew, Mapping):
                        if tuple(crew) != clan._fields:  # fields is mark
                            raise InvalidValueError(f"Mismatching fields clan="
                                                    f"{clan._fields} and keys crew="
                                                    f"{tuple(crew)}.")

                        crew = clan(**crew)  # convert to clan

                    elif isinstance(crew, nonStringSequence):
                        crew = clan(*crew)  # convert to clan assumes elements in correct order

                    else:
                        raise InvalidValueError(f"Invalid {crew=}.")

                data = clan(*(cstg.kls(**{cstg.prm if cstg.prm is not None else 'qb64': val})
                              for cstg, val in zip(cast, crew)))
                # data = clan(*(klas(qb64=val) for klas, val in zip(cast, crew)))

            elif qb64:
                if hasattr(qb64, "encode"):
                    qb64 = qb64.encode()

                if strip:
                    if not isinstance(qb64, bytearray):
                        qb64 = bytearray(qb64)

                    data = clan(*(cstg.kls(qb64b=qb64, strip=strip) for cstg in cast))

                else:
                    o = 0  # offset into memoryview of qb64
                    pis = []  # primitive instances
                    mv = memoryview(qb64)
                    for cstg in cast:  # Castage
                        pi = cstg.kls(qb64b=mv[o:])
                        pis.append(pi)
                        o += len(pi.qb64b)
                    data = clan(*pis)

            elif qb2:
                if strip:
                    if not isinstance(qb2, bytearray):
                        qb2 = bytearray(qb2)

                    data = clan(*(cstg.kls(qb2=qb2, strip=strip) for cstg in cast))

                else:
                    o = 0  # offset into memoryview of qb2
                    pis = []  # primitive instances
                    mv = memoryview(qb2)
                    for cstg in cast:  # Castage
                        pi = cstg.kls(qb2=mv[o:])
                        pis.append(pi)
                        o += len(pi.qb2)
                    data = clan(*pis)

            else:
                raise EmptyMaterialError("Need crew or qb64 or qb2.")


        self._data = data
        self._cast = (cast if cast is not None else
                      self.clan(*(Castage(val.__class__) for val in self.data)))




    @property
    def data(self):
        """Returns:
            data (NamedTuple): ._data namedtuple of primitive instances

        Getter for ._data makes it read only
        """
        return self._data


    @property
    def clan(self):
        """Returns:
              clan (type[NamedTuple]): class of .data

        """
        return self.data.__class__

    @property
    def name(self):
        """Returns:
              name (str): name of class of .data

        """
        return self.data.__class__.__name__

    @property
    def cast(self):
        """Return:
            cast (NamedTuple): named primitive classes in .data

        Getter for ._cast makes it read only when not None
        """
        return self._cast

    @property
    def crew(self):
        """Return:
            crew (NamedTuple): named qb64 field values from .data

        """
        return (self.clan(*(getattr(val, cstg.prm if cstg.prm is not None else "qb64")
                    for cstg, val in zip(self.cast, self.data))))
        # return self.clan(*(val.qb64 for val in self.data))


    @property
    def asdict(self):
        """Returns:
            map (dict): .data as a dictionary

        """
        return self.data._asdict()


    @property
    def qb64(self):
        """Returns:
              qb64 (str): concatenated qb64 of each primitive in .data
        """
        return (''.join(val.qb64 for val in self.data))


    @property
    def qb64b(self):
        """Returns:
              qb64b (bytes): concatenated qb64b of each primitive in .data
        """
        return (b''.join(val.qb64b for val in self.data))


    @property
    def qb2(self):
        """Returns:
              qb2 (bytes): concatenated qb2 of each primitive in .data

        """
        return (b''.join(val.qb2 for val in self.data))




class Sealer(Structor):
    """Sealer is Structor subclass each instance holds a namedtuple .data of
    named values belonging to KERI Seals for anchoring in messages or adding
    to message attachments.

    See Structor class for more details.


    Inherited Class Attributes:
        Clans (type[Namedtuple]): each value is known NamedTuple class keyed
            by its own field names (tuple). Enables easy query of its values() to
            find known data types given field names tuple.

        Casts (NamedTuple): each value is primitive class of cast keyed by fields
            names of the associated NamedTuple class in .Clans. Enables finding
            known primitive classes given NamedTuple class of clan or instance
            of cast or crew.

    When known casts or provided in .Clans/.Casts then more flexible creation
    is supported for different types of provided cast and crew.
    When no clan is provided and an unknown cast and/or crew are provided as
    Mappings then Structor may create custom clan from the names given by the
    cast and/or crew keys(). Subclasses may override this behavior by raising
    an exception for unknown or custom clans.


    Inherited Properties:
        data (NamedTuple): fields are named instances of CESR primitives
        clan (type[NamedTuple]): class reference of .data's class
        cast (NamedTuple): CESR primitive class references of .data's primitive
                           instances
        crew (NamedTuple): named qb64 values of .data's primitive instances
        qb64 (str): concatenated data values as qb64 str of data's primitives
        qb64b (bytes): concatenated data values as qb64b  of data's primitives
        qb2 (bytes): concatenated data values as qb2 bytes of data's primitives


    Methods:


    Hidden:
        _data (NamedTuple): named CESR primitive instances


    """
    Clans = SClanDom  # known namedtuple clans. Override in subclass with non-empty
    Casts = SCastDom  # known namedtuple casts. Override in subclass with non-empty
    # Create .Names dict that maps clan/cast fields names to its namedtuple
    # class type name so can look up a know clan or cast given a matching set
    # of either field names from a namedtuple or keys from a dict.
    Names = {tuple(clan._fields): clan.__name__ for clan in Clans}


    def __init__(self, *pa, **kwa):
        """Initialize instance


        Inherited Parameters:
            data (NamedTuple): fields are named primitive instances for .data
                Given data can derive clan, cast, crew, qb64, and qb2
            clan (type[NamedTuple]): provides class reference for generated .data
                when data missing.
            cast (NamedTuple | dict | Iterable): each value provides CESR
                primitive subclass reference used to create primitive instances
                for generating .data. Can be used to infer namedtuple type of
                .data when data and clan missing. Takes precendence over crew.
            crew (NamedTuple | dict | Iterable): each value provides qb64 value
                of primitive for generating .data with .cast when data missing.
                Can be used to infer namedtuple type of .data when data and clan
                missing.
            qb64 (str | bytes | bytearray): concatenation of qb64 data values to
                generate .data with data and crew missing.
            qb2 (bytes | bytearray): concatenation of qb2 data values to generate
                .data when data and crew and qb64 missing.
            strip (bool): False means do not strip each value from qb64 or qb2.
                            Default is False.
                          True means if qb64 or qb2 are bytearray then strip
                            contained concatenated data values. Enables parser
                            to extract data fields from front of CESR stream.

        """

        super(Sealer, self).__init__(*pa, **kwa)



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""
import argparse
import json
import sys

import qrcode
from hio import help
from hio.base import doing
from hio.help import Hict

from keri.app import httping, connecting
from keri.app.agenting import httpClient
from keri.app.cli.common import existing
from keri.app.httping import CESR_DESTINATION_HEADER
from keri.core import coring

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Perform authentication against an witness to get a OTP code')
parser.set_defaults(handler=lambda args: auth(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--alias', '-a', help='human readable alias for the identifier to whom the credential was issued',
                    required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument("--witness", '-w', help="the witness AID or alias to authenticate against", required=True)
parser.add_argument("--url-only", '-u', dest="url", help="display only the URL (no QR Code).", required=False,
                    action="store_true")


def auth(args):
    """ Command line handler for authenticating  against a witness by retrieving the secret or a TOTP

    Parameters:
        args(Namespace): parsed command line arguments

    """

    ed = AuthDoer(name=args.name,
                  alias=args.alias,
                  base=args.base,
                  bran=args.bran,
                  witness=args.witness,
                  urlOnly=args.url)
    return [ed]


class AuthDoer(doing.DoDoer):

    def __init__(self, name, alias, base, bran, witness, urlOnly):
        self.hby = existing.setupHby(name=name, base=base, bran=bran)
        self.hab = self.hby.habByName(alias)
        self.org = connecting.Organizer(hby=self.hby)
        self.urlOnly = urlOnly

        if witness in self.hby.kevers:
            wit = witness
        else:
            wit = self.org.find("alias", witness)
            if len(wit) != 1:
                raise ValueError(f"invalid recipient {witness}")
            wit = wit[0]['id']

        if not wit:
            raise ValueError(f"unknown witness {witness}")

        self.witness = wit
        self.clienter = httping.Clienter()
        doers = [doing.doify(self.authDo), self.clienter]

        super(AuthDoer, self).__init__(doers=doers)

    def authDo(self, tymth, tock=0.0):
        """ Export credential from store and any related material

        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock (float): injected initial tock value

        Returns:  doifiable Doist compatible generator method

        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        body = bytearray()
        for msg in self.hab.db.clonePreIter(pre=self.hab.pre):
            body.extend(msg)

        fargs = dict([("kel", body.decode("utf-8"))])

        if self.hab.kever.delegated:
            delkel = bytearray()
            for msg in self.hab.db.clonePreIter(self.hab.kever.delpre):
                delkel.extend(msg)

            fargs['delkel'] = delkel.decode("utf-8")

        headers = (Hict([
            ("Content-Type", "multipart/form-data"),
            (CESR_DESTINATION_HEADER, self.witness)
        ]))

        client, clientDoer = httpClient(self.hab, self.witness)
        self.extend([clientDoer])

        client.request(
            method="POST",
            path=f"{client.requester.path}/aids",
            headers=headers,
            fargs=fargs
        )
        while not client.responses:
            yield self.tock

        rep = client.respond()
        if rep.status == 200:
            data = json.loads(rep.body)

            totp = data["totp"]
            m = coring.Matter(qb64=totp)  # refactor this to use cipher
            d = coring.Matter(qb64=self.hab.decrypt(ser=m.raw))
            otpurl = f"otpauth://totp/KERIpy:{self.witness}?secret={d.raw.decode('utf-8')}&issuer=KERIpy"

            if not self.urlOnly:
                qr = qrcode.QRCode()
                qr.add_data(otpurl)

                qr.print_ascii()

            print(otpurl)

        else:
            raise ValueError(rep.body)

        self.remove([clientDoer, self.clienter])



# -*- encoding: utf-8 -*-
"""
KERI
keri.vdr.eventing module

VC TEL  support
"""

import json
import logging
from dataclasses import asdict
from math import ceil
from  ordered_set import OrderedSet as oset

from hio.help import decking

from keri import kering
from .. import core
from .. import help
from ..core import serdering, coring, indexing
from ..core.coring import (MtrDex, Kinds, versify, Prefixer,
                           Ilks, Seqner, Verfer, Number)
from ..core.signing import (Salter,)
from ..core.eventing import SealEvent, ample, TraitDex, verifySigs
from ..db import basing, dbing
from ..db.dbing import dgKey, snKey, splitSnKey
from ..help import helping
from ..kering import (MissingWitnessSignatureError, Version,
                      MissingAnchorError, ValidationError, OutOfOrderError, LikelyDuplicitousError)
from ..vdr import viring

logger = help.ogler.getLogger()


def incept(
        pre,
        toad=None,
        baks=None,
        nonce=None,
        cnfg=None,
        version=Version,
        kind=Kinds.json,
        code=MtrDex.Blake3_256,
):
    """ Returns serder of credential registry inception (vcp) message event

    Returns serder of vcp message event
    Utility function to create a Registry inception event

    Parameters:
         pre (str): issuer identifier prefix qb64
         toad (Union(int,str)): int or str hex of backer threshold
         baks (list): the initial list of backers prefixes for VCs in the Registry
         nonce (str): qb64 encoded ed25519 random seed of credential registry
         cnfg (list): is list of strings TraitDex of configuration traits

         version (Versionage): the API version
         kind (str): the event type
         code (str): default code for Prefixer

    Returns:
        Serder: Event message Serder

    """

    vs = versify(version=version, kind=kind, size=0)
    isn = 0
    ilk = Ilks.vcp

    cnfg = cnfg if cnfg is not None else []

    baks = baks if baks is not None else []
    if TraitDex.NoBackers in cnfg and len(baks) > 0:
        raise ValueError("{} backers specified for NB vcp, 0 allowed".format(len(baks)))

    if len(oset(baks)) != len(baks):
        raise ValueError("Invalid baks = {}, has duplicates.".format(baks))

    if isinstance(toad, str):
        toad = int(toad, 16)
    elif toad is None:
        if not baks:
            toad = 0
        else:  # compute default f and m for len(baks)
            toad = ample(len(baks))

    if baks:
        if toad < 1 or toad > len(baks):  # out of bounds toad
            raise ValueError("Invalid toad = {} for baks = {}".format(toad, baks))
    else:
        if toad != 0:  # invalid toad
            raise ValueError("Invalid toad = {} for baks = {}".format(toad, baks))

    nonce = nonce if nonce is not None else Salter().qb64
    ked = dict(v=vs,  # version string
               t=ilk,
               d="",
               i="",  # qb64 prefix
               ii=pre,
               s="{:x}".format(isn),  # hex string no leading zeros lowercase
               c=cnfg,
               bt="{:x}".format(toad),  # hex string no leading zeros lowercase
               b=baks,  # list of qb64 may be empty
               n=nonce  # nonce of random bytes to make each registry unique
               )

    serder = serdering.SerderKERI(sad=ked, makify=True)
    return serder


def rotate(
        regk,
        dig,
        sn=1,
        toad=None,
        baks=None,
        cuts=None,
        adds=None,
        version=Version,
        kind=Kinds.json,
):
    """ Returns serder of registry rotation (brt) message event

    Returns serder of vrt message event
    Utility function to create a Registry rotation event

    Parameters:
        regk (str): identifier prefix qb64
        dig (str): qb64 digest or prior event
        sn (int): sequence number
        toad (int): int or str hex of witness threshold
        baks (list): prior backers prefixes qb64
        cuts (list): witness prefixes to cut qb64
        adds (list): witness prefixes to add qb64
        version (Versionage): the API version
        kind (str): the event type

    Returns:
        Serder: event message Serder

    """

    if sn < 1:
        raise ValueError("Invalid sn = {} for vrt.".format(sn))

    vs = versify(version=version, kind=kind, size=0)
    ilk = Ilks.vrt

    baks = baks if baks is not None else []
    bakset = oset(baks)
    if len(bakset) != len(baks):
        raise ValueError("Invalid baks = {}, has duplicates.".format(baks))

    cuts = cuts if cuts is not None else []
    cutset = oset(cuts)
    if len(cutset) != len(cuts):
        raise ValueError("Invalid cuts = {}, has duplicates.".format(cuts))

    if (bakset & cutset) != cutset:  # some cuts not in wits
        raise ValueError("Invalid cuts = {}, not all members in baks.".format(cuts))

    adds = adds if adds is not None else []
    addset = oset(adds)
    if len(addset) != len(adds):
        raise ValueError("Invalid adds = {}, has duplicates.".format(adds))

    if cutset & addset:  # non empty intersection
        raise ValueError("Intersecting cuts = {} and  adds = {}.".format(cuts, adds))

    if bakset & addset:  # non empty intersection
        raise ValueError("Intersecting baks = {} and  adds = {}.".format(baks, adds))

    newbakset = (bakset - cutset) | addset

    if len(newbakset) != (len(baks) - len(cuts) + len(adds)):  # redundant?
        raise ValueError("Invalid member combination among baks = {}, cuts ={}, "
                         "and adds = {}.".format(baks, cuts, adds))

    if isinstance(toad, str):
        toad = int(toad, 16)
    elif toad is None:
        if not newbakset:
            toad = 0
        else:  # compute default f and m for len(newbakset)
            toad = ample(len(newbakset))

    if newbakset:
        if toad < 1 or toad > len(newbakset):  # out of bounds toad
            raise ValueError("Invalid toad = {} for resultant wits = {}"
                             "".format(toad, list(newbakset)))
    else:
        if toad != 0:  # invalid toad
            raise ValueError("Invalid toad = {} for resultant wits = {}"
                             "".format(toad, list(newbakset)))

    ked = dict(v=vs,  # version string
               t=ilk,
               d="",
               i=regk,  # qb64 prefix
               p=dig,
               s="{:x}".format(sn),  # hex string no leading zeros lowercase
               bt="{:x}".format(toad),  # hex string no leading zeros lowercase
               br=cuts,  # list of qb64 may be empty
               ba=adds,  # list of qb64 may be empty
               )

    serder = serdering.SerderKERI(sad=ked, makify=True)
    return serder


def issue(
        vcdig,
        regk,
        version=Version,
        kind=Kinds.json,
        dt=None
):
    """ Returns serder of issuance (iss) message event

    Returns serder of iss message event
    Utility function to create a VC issuance event

    Parameters:
        vcdig (str): qb64 SAID of credential
        regk (str): qb64 AID of credential registry
        version (Versionage): the API version
        kind (str): the event type
        dt (str): ISO 8601 formatted date string of issuance date

    Returns:
        Serder: event message Serder

    """

    vs = versify(version=version, kind=kind, size=0)
    ked = dict(v=vs,  # version string
               t=Ilks.iss,
               d="",
               i=vcdig,  # qb64 prefix
               s="{:x}".format(0),  # hex string no leading zeros lowercase
               ri=regk,
               dt=helping.nowIso8601()
               )
    if dt is not None:
        ked["dt"] = dt

    serder = serdering.SerderKERI(sad=ked, makify=True)
    return serder


def revoke(
        vcdig,
        regk,
        dig,
        version=Version,
        kind=Kinds.json,
        dt=None
):
    """ Returns serder of backerless credential revocation (rev) message event

    Returns serder of rev message event
    Utility function to create a VC revocation vent

    Parameters:
        vcdig (str): qb64 SAID of credential
        regk (str): qb64 AID of credential registry
        dig (str): digest of previous event qb64
        version (Versionage): the API version
        kind (str): the event type
        dt (str): ISO 8601 formatted date string of revocation date

    Returns:
        Serder: event message Serder

    """

    vs = versify(version=version, kind=kind, size=0)
    isn = 1
    ilk = Ilks.rev

    ked = dict(v=vs,
               t=ilk,
               d="",
               i=vcdig,
               s="{:x}".format(isn),  # hex string no leading zeros lowercase
               ri=regk,
               p=dig,
               dt=helping.nowIso8601()
               )

    if dt is not None:
        ked["dt"] = dt

    _, ked = coring.Saider.saidify(sad=ked)

    serder = serdering.SerderKERI(sad=ked, makify=True)
    return serder


def backerIssue(
        vcdig,
        regk,
        regsn,
        regd,
        version=Version,
        kind=Kinds.json,
        dt=None,
):
    """ Returns serder of backer issuance (bis) message event

    Returns serder of bis message event
    Utility function to create a VC issuance event

    Parameters:
        vcdig (str): qb64 SAID of credential
        regk (str): qb64 AID of credential registry
        regsn (int): sequence number of anchoring registry TEL event
        regd (str): digest qb64 of anchoring registry TEL event
        version (Versionage): the API version
        kind (str): the event type
        dt (str): ISO 8601 formatted date string of issuance date

    Returns:
        Serder: event message Serder

    """

    vs = versify(version=version, kind=kind, size=0)
    isn = 0
    ilk = Ilks.bis

    seal = SealEvent(regk, "{:x}".format(regsn), regd)

    ked = dict(v=vs,  # version string
               t=ilk,
               d="",
               i=vcdig,  # qb64 prefix
               ii=regk,
               s="{:x}".format(isn),  # hex string no leading zeros lowercase
               ra=seal._asdict(),
               dt=helping.nowIso8601(),
               )
    _, ked = coring.Saider.saidify(sad=ked)

    if dt is not None:
        ked["dt"] = dt

    serder = serdering.SerderKERI(sad=ked, makify=True)
    return serder


def backerRevoke(
        vcdig,
        regk,
        regsn,
        regd,
        dig,
        version=Version,
        kind=Kinds.json,
        dt=None
):
    """ Returns serder of backer credential revocation (brv) message event

    Returns serder of brv message event
    Utility function to create a VC revocation event

    Parameters:
        vcdig (str): qb64 SAID of credential
        regk (str): qb64 AID of credential registry
        regsn (int): sequence number of anchoring registry TEL event
        regd (str): digest qb64 of anchoring registry TEL event
        dig (str) digest of previous event qb64
        version (Versionage): the API version
        kind (str): the event type
        dt (str): ISO 8601 formatted date string of issuance date

    Returns:
        Serder: event message Serder

    """

    vs = versify(version=version, kind=kind, size=0)
    isn = 1
    ilk = Ilks.brv

    seal = SealEvent(regk, "{:x}".format(regsn), regd)

    ked = dict(v=vs,
               t=ilk,
               d="",
               i=vcdig,
               s="{:x}".format(isn),  # hex string no leading zeros lowercase
               p=dig,
               ra=seal._asdict(),
               dt=helping.nowIso8601(),
               )
    _, ked = coring.Saider.saidify(sad=ked)

    if dt is not None:
        ked["dt"] = dt

    serder = serdering.SerderKERI(sad=ked, makify=True)
    return serder


def state(pre,
          said,
          sn,
          ri,
          eilk,
          dts=None,  # default current datetime
          toad=None,  # default based on wits
          wits=None,  # default to []
          cnfg=None,  # default to []
          version=Version,
          ):
    """
    Utility function to create a RegStateRecord of state notice of a given
        Registry Event Log (REL)

        Returns:
            rsr: (RegStateRecord): instance

    Parameters:
        pre (str): identifier prefix qb64
        sn (int): int sequence number of latest event
        said (str): digest of latest event
        ri (str): qb64 AID of credential registry
        eilk (str): message type (ilk) oflatest event
        a (dict): key event anchored seal data
        dts (str) ISO 8601 formated current datetime
        toad (int): int of witness threshold
        wits (list): list of witness prefixes qb64
        cnfg (list): list of strings TraitDex of configuration traits
        version (str): Version instance
        kind (str): serialization kind

    Returns:
        Serder: Event message Serder

    Key State Dict
    {
        "v": "KERI10JSON00011c_",
        "i": "EaU6JR2nmwyZ-i0d8JZAoTNZH3ULvYAfSVPzhzS6b5CM",
        "s": "2":,
        "p": "EYAfSVPzhzZ-i0d8JZS6b5CMAoTNZH3ULvaU6JR2nmwy",
        "d": "EAoTNZH3ULvaU6JR2nmwyYAfSVPzhzZ-i0d8JZS6b5CM",
        "ri": "EYAfSVPzhzZ-i0d8JZS6b5CMAoTNZH3ULvaU6JR2nmwy",
        "dt": "2020-08-22T20:35:06.687702+00:00",
        "et": "vrt",
        "a": {i=12, d="EYAfSVPzhzS6b5CMaU6JR2nmwyZ-i0d8JZAoTNZH3ULv"},
        "k": ["DaU6JR2nmwyZ-i0d8JZAoTNZH3ULvYAfSVPzhzS6b5CM"],
        "n": "EZ-i0d8JZAoTNZH3ULvaU6JR2nmwyYAfSVPzhzS6b5CM",
        "bt": "1",
        "b": ["DnmwyYAfSVPzhzS6b5CMZ-i0d8JZAoTNZH3ULvaU6JR2"],
        "di": "EYAfSVPzhzS6b5CMaU6JR2nmwyZ-i0d8JZAoTNZH3ULv",
        "c": ["EO"],
    }

    """
    #vs = versify(version=version, kind=kind, size=0)

    if sn < 0:
        raise ValueError("Negative sn = {} in key state.".format(sn))

    if eilk not in (Ilks.vcp, Ilks.vrt):
        raise ValueError("Invalid evernt type et=  in key state.".format(eilk))

    if dts is None:
        dts = helping.nowIso8601()

    wits = wits if wits is not None else []
    witset = oset(wits)
    if len(witset) != len(wits):
        raise ValueError("Invalid wits = {}, has duplicates.".format(wits))

    if toad is None:
        if not witset:
            toad = 0
        else:
            toad = max(1, ceil(len(witset) / 2))

    if witset:
        if toad < 1 or toad > len(witset):  # out of bounds toad
            raise ValueError("Invalid toad = {} for resultant wits = {}"
                             "".format(toad, list(witset)))
    else:
        if toad != 0:  # invalid toad
            raise ValueError("Invalid toad = {} for resultant wits = {}"
                             "".format(toad, list(witset)))

    cnfg = cnfg if cnfg is not None else []

    rsr = viring.RegStateRecord(
               vn=list(version),  # version number as list [major, minor]
               i=ri,  # qb64 registry SAID
               s="{:x}".format(sn),  # lowercase hex string no leading zeros
               d=said,
               ii=pre,
               dt=dts,
               et=eilk,
               bt="{:x}".format(toad),  # hex string no leading zeros lowercase
               b=wits,  # list of qb64 may be empty
               c=cnfg if cnfg is not None else [],
               )
    return rsr  # return RegStateRecord  use asdict(rsr) to get dict version


def vcstate(vcpre,
            said,
            sn,
            ri,
            eilk,
            a,
            ra=None,
            dts=None,  # default current datetime
            version=Version,
            kind=Kinds.json,
            ):
    """ Returns the credential transaction state notification

    Returns serder of credential transaction state notification message.
    Utility function to automate creation of tsn events.

    Parameters:
        vcpre (str): is qb64 SAID of the credential
        said (str): is qb64 digest of latest event
        sn (int): sequence number of latest event
        ri (str): registry identifier
        ra (dict): optional registry seal for registries with backers
        eilk (str): is message type (ilk) of latest event
        a (dict): is seal for anchor in KEL
        dts (str): iso8601 formatted date string of state
        version (Version): is KERI version instance
        kind (str): is serialization kind

    Credential Transaction State Dict
    {
       "v": "KERI10JSON00012d_",
       "i": "EDGhJ8V1tuwH55Bk0fBFe9L0za2BUNOt2FX4GUeOLNHQ",
       "s": "0",
       "d": "ENNTabgWbaNqOKLqEZdQCjxbafwwSoXNzAsE1Enq-kdk",
       "ri": "EoN_Ln_JpgqsIys-jDOH8oWdxgWqs7hzkDGeLWHb9vSY",
       "a": {
        "s": 3,
        "d": "Ex7i6wv4YzDRTO9_iHkTQSXrvLYldSd_UEjNfqia3Pqc"
       },
       "dt": "2021-01-01T00:00:00.000000+00:00",
       "et": "bis"
    }
    """

    if sn < 0:
        raise ValueError("Negative sn = {} in key state.".format(sn))

    if eilk not in (Ilks.iss, Ilks.bis, Ilks.rev, Ilks.brv):
        raise ValueError("Invalid event type et=  in key state.".format(eilk))

    if dts is None:
        dts = helping.nowIso8601()

    if ra is None:
        ra = dict()

    vsr = viring.VcStateRecord(vn=list(version),  # version string
                               i=vcpre,  # qb64 prefix
                               s="{:x}".format(sn),  # lowercase hex string no leading zeros
                               d=said,
                               ri=ri,
                               ra=ra,
                               a=a,
                               dt=dts,
                               et=eilk,
                               )

    return vsr  # return vc state record data class


def query(regk,
          vcid,
          route="",
          replyRoute="",
          dt=None,
          dta=None,
          dtb=None,
          stamp=None,
          version=Version,
          kind=Kinds.json
          ):
    """ Returns serder of credentialquery (qry) event message.

    Returns serder of query event message.
    Utility function to automate creation of interaction events.

    Parameters:
        regk (str): qb64 AID of credential registry
        vcid (str): qb64 SAID of credential
        route (str): namesapaced path, '/' delimited, that indicates data flow
                     handler (behavior) to processs the query
        replyRoute (str): namesapaced path, '/' delimited, that indicates data flow
                     handler (behavior) to processs reply message to query if any.
        dt (str): ISO 8601 formatted datetime query
        dta (str): ISO 8601 formatted datetime after query
        dtb (str): ISO 8601 formatted datetime before query
        stamp (str): ISO 8601 formatted current datetime of query message
        version (Versionage): the API version
        kind (str): the event type

    Returns:
        Serder: query event message Serder

    """
    qry = dict(i=vcid, ri=regk)

    if dt is not None:
        qry["dt"] = dt

    if dta is not None:
        qry["dta"] = dt

    if dtb is not None:
        qry["dtb"] = dt

    return core.eventing.query(route=route,
                               replyRoute=replyRoute,
                               query=qry,
                               stamp=stamp,
                               version=version,
                               kind=kind)


class Tever:
    """
    Tever is KERI/ACDC transaction event log verifier class
    Only supports current version VERSION

    Has the following public attributes and properties:

    Class Attributes:
        .NoRegistrarBackers is Boolean
                True means do not allow backers (default to witnesses of controlling KEL)
                False means allow backers (ignore witnesses of controlling KEL)

    Attributes:
        .db is reference to Baser instance that managers the LMDB database
        .reg is regerence to Registry instance that manages VC LMDB database
        .regk is fully qualified base64 identifier prefix of own Registry if any
        .local is Boolean
            True means only process msgs for own events if .regk
            False means only process msgs for not own events if .regk
        .version is version of current event state
        .prefixer is prefixer instance fParemtersor current event state
        .sn is sequence number int
        .serder is Serder instance of current event with .serder.diger for digest
        .toad is int threshold of accountable duplicity
        .baks is list of qualified qb64 aids for backers
        .cuts is list of qualified qb64 aids for backers cut from prev wits list
        .adds is list of qualified qb64 aids for backers added to prev wits list
        .noBackers is boolean trait True means do not allow backers

    """
    NoRegistrarBackers = False

    def __init__(self, cues=None, rsr=None, serder=None, seqner=None, saider=None,
                 bigers=None, db=None, reger=None, noBackers=None, estOnly=None,
                 regk=None, local=False):
        """ Create incepting tever and state from registry inception serder

        Create incepting tever and state from registry inception serder

        Parameters:
            serder (Serder): instance of registry inception event
            rsr (RegStateRecord): transaction state notice state message Serder
            seqner (Seqner): issuing event sequence number from controlling KEL.
            saider (Saider): issuing event said from controlling KEL.
            bigers (list): list of Siger instances of indexed backer signatures of
                event. Index is offset into baks list of latest est event
            db (Baser): instance of baser lmdb database
            reger (Reger): instance of VC lmdb database
            noBackers (bool): True means do not allow backer configuration
            estOnly (bool): True means do not allow interaction events
            regk (str): identifier prefix of own or local registry. May not be the
                prefix of this Tever's event. Some restrictions if present
            local (bool): True means only process msgs for own controller's
                events if .regk. False means only process msgs for not own events
                if .regk

        Returns:
            Tever:  instance representing credential Registry

        """

        if not (rsr or serder):
            raise ValueError("Missing required arguments. Need state or serder")

        self.reger = reger if reger is not None else viring.Reger()
        self.cues = cues if cues is not None else decking.Deck()

        self.db = db if db is not None else basing.Baser(reopen=True)
        self.local = True if local else False

        if rsr:  # preload from state
            self.reload(rsr)
            return

        self.version = serder.version
        self.regk = regk

        ilk = serder.ked["t"]
        if ilk not in [Ilks.vcp]:
            raise ValidationError("Expected ilk {} got {} for evt: {}".format(Ilks.vcp, ilk, serder))

        self.ilk = ilk
        self.incept(serder=serder)
        self.config(serder=serder, noBackers=noBackers, estOnly=estOnly)

        bigers = self.valAnchorBigs(serder=serder,
                                    seqner=seqner,
                                    saider=saider,
                                    bigers=bigers,
                                    toad=self.toad,
                                    baks=self.baks)

        self.logEvent(pre=self.prefixer.qb64b,
                      sn=0,
                      serder=serder,
                      seqner=seqner,
                      saider=saider,
                      bigers=bigers,
                      baks=self.baks)

        self.regk = self.prefixer.qb64

    def reload(self, rsr):
        """ Reload Tever attributes (aka its state) from state serder

        Reload Tever attributes (aka its state) from state serder

        Parameters:
            rsr (RegStateRecord): instance of key stat notice 'ksn' message body

        """

        ked = asdict(rsr)

        self.version = rsr.vn
        self.pre = ked["ii"]
        self.regk = ked["i"]
        self.prefixer = Prefixer(qb64=self.regk)
        self.sn = int(ked['s'], 16)
        self.ilk = ked["et"]
        self.toad = int(ked["bt"], 16)
        self.baks = ked["b"]

        self.noBackers = True if TraitDex.NoBackers in ked["c"] else False
        self.estOnly = True if TraitDex.EstOnly in ked["c"] else False

        if (raw := self.reger.getTvt(key=dgKey(pre=self.prefixer.qb64,
                                               dig=ked['d']))) is None:
            raise kering.MissingEntryError("Corresponding event for state={} not found."
                                           "".format(ked))
        self.serder = serdering.SerderKERI(raw=bytes(raw))

    def state(self):  #state(self, kind=Serials.json)
        """ Returns RegStateRecord of state notice of given Registry Event Log
        (REL)

        Returns:
            rsr: (RegStateRecord): instance for this Tever


        """

        cnfg = []
        if self.noBackers:
            cnfg.append(TraitDex.NoBackers)

        dgkey = dbing.dgKey(self.regk, self.serder.said)
        couple = self.reger.getAnc(dgkey)
        ancb = bytearray(couple)
        seqner = coring.Seqner(qb64b=ancb, strip=True)
        diger = coring.Diger(qb64b=ancb, strip=True)

        return (state(pre=self.pre,
                      said=self.serder.said,
                      sn=self.sn,
                      ri=self.regk,
                      dts=None,
                      eilk=self.ilk,
                      #a=dict(s=seqner.sn, d=diger.qb64),
                      toad=self.toad,
                      wits=self.baks,
                      cnfg=cnfg,
                      #kind=kind
                      )
                )

    def incept(self, serder):
        """  Validate registry inception event and initialize local attributes

        Parse and validate registry inception event for this Tever.  Update all
        local attributes with initial values.

        Parameters:
            serder (Serder): registry inception event (vcp)

        """

        ked = serder.ked
        self.pre = ked["ii"]  # which is not the AID of the serder in ked["i"]
        self.prefixer = Prefixer(qb64=serder.pre)  # this not related to self.pre
        #if not self.prefixer.verify(ked=ked, prefixed=True):  # invalid prefix
            #raise ValidationError("Invalid prefix = {} for registry inception evt = {}."
                                  #.format(self.prefixer.qb64, ked))


        self.sn = Number(numh=ked["s"]).validate(inceptive=True).sn

        self.cuts = []  # always empty at inception since no prev event
        self.adds = []  # always empty at inception since no prev event
        baks = ked["b"]
        if len(oset(baks)) != len(baks):
            raise ValidationError("Invalid baks = {}, has duplicates for evt = {}."
                                  "".format(baks, ked))
        self.baks = baks

        toad = int(ked["bt"], 16)
        if baks:
            if toad < 1 or toad > len(baks):  # out of bounds toad
                raise ValidationError("Invalid toad = {} for baks = {} for evt = {}."
                                      "".format(toad, baks, ked))
        else:
            if toad != 0:  # invalid toad
                raise ValidationError("Invalid toad = {} for baks = {} for evt = {}."
                                      "".format(toad, baks, ked))
        self.toad = toad
        self.serder = serder

    def config(self, serder, noBackers=None, estOnly=None):
        """ Process cnfg field for configuration traits

        Parse and validate the configuration options for registry inception from
        the `c` field of the provided inception event.

        Parameters:
            serder (Serder): credential registry inception event `vcp`
            noBackers (bool): override flag for specifying a registry with no additional backers
                              beyond the controlling KEL's witnesses


        """
        # assign traits
        self.noBackers = (True if (noBackers if noBackers is not None
                                   else self.NoRegistrarBackers)
                          else False)  # ensure default noBackers is boolean

        self.estOnly = (True if (estOnly if estOnly is not None
                                   else False)
                          else False)  # ensure default estOnly is boolean

        cnfg = serder.ked["c"]  # process cnfg for traits
        if TraitDex.NoBackers in cnfg:
            self.noBackers = True
        if TraitDex.EstOnly in cnfg:
            self.estOnly = True

    def update(self, serder, seqner=None, saider=None, bigers=None):
        """ Process registry non-inception events.

        Process non-inception registry and credential events and update local
        Tever state for registry or credential

        Parameters:
            serder (Serder): instance of issuance or backer issuance event
            seqner (Seqner): issuing event sequence number from controlling KEL.
            saider (Saider): issuing event SAID from controlling KEL.
            bigers (list): of Siger instances of indexed witness signatures.
                Index is offset into wits list of associated witness nontrans pre
                from which public key may be derived.

        """

        ked = serder.ked
        ilk = ked["t"]
        #sn = ked["s"]

        icp = ilk in (Ilks.iss, Ilks.bis)

        # validate SN for
        #sn = validateSN(sn, inceptive=icp)
        sn = Number(numh=ked["s"]).validate(inceptive=icp).sn

        if ilk in (Ilks.vrt,):
            if self.noBackers is True:
                raise ValidationError("invalid rotation evt {} against backerless registry {}".
                                      format(ked, self.regk))

            toad, baks, cuts, adds = self.rotate(serder, sn=sn)

            bigers = self.valAnchorBigs(serder=serder,
                                        seqner=seqner,
                                        saider=saider,
                                        bigers=bigers,
                                        toad=toad,
                                        baks=baks)

            self.sn = sn
            self.serder = serder
            self.ilk = ilk
            self.toad = toad
            self.baks = baks
            self.cuts = cuts
            self.adds = adds

            self.logEvent(pre=self.prefixer.qb64b,
                          sn=sn,
                          serder=serder,
                          seqner=seqner,
                          saider=saider,
                          bigers=bigers,
                          baks=self.baks)

            return

        elif ilk in (Ilks.iss, Ilks.bis):
            self.issue(serder, seqner=seqner, saider=saider, sn=sn, bigers=bigers)
        elif ilk in (Ilks.rev, Ilks.brv):
            self.revoke(serder, seqner=seqner, saider=saider, sn=sn, bigers=bigers)
        else:  # unsupported event ilk so discard
            raise ValidationError("Unsupported ilk = {} for evt = {}.".format(ilk, ked))


    def rotate(self, serder, sn):
        """ Process registry management TEL, non-inception events (vrt)

        Parameters:
            serder (Serder): registry rotation event
            sn (int): sequence number of event

        Returns:
            int: calculated backer threshold
            list: new list of backers after applying cuts and adds to previous list
            list: list of backer adds processed from event
            list: list of backer cuts processed from event

        """

        ked = serder.ked
        ilk = ked["t"]
        dig = ked["p"]


        #labels = VRT_LABELS  # assumes ilk == Ilks.vrt
        #for k in labels:
            #if k not in ked:
                #raise ValidationError("Missing element = {} from {} event for "
                                      #"evt = {}.".format(k, ilk, ked))

        if serder.pre != self.prefixer.qb64:
            raise ValidationError("Mismatch event aid prefix = {} expecting"
                                  " = {} for evt = {}.".format(ked["i"],
                                                               self.prefixer.qb64,
                                                               ked))
        if not sn == (self.sn + 1):  # sn not in order
            raise ValidationError("Invalid sn = {} expecting = {} for evt "
                                  "= {}.".format(sn, self.sn + 1, ked))

        if not self.serder.compare(said=dig):  # prior event dig not match
            raise ValidationError("Mismatch event dig = {} with state dig"
                                  " = {} for evt = {}.".format(ked["p"],
                                                               self.serder.said,
                                                               ked))

        witset = oset(self.baks)
        cuts = ked["br"]
        cutset = oset(cuts)
        if len(cutset) != len(cuts):
            raise ValidationError("Invalid cuts = {}, has duplicates for evt = "
                                  "{}.".format(cuts, ked))

        if (witset & cutset) != cutset:  # some cuts not in baks
            raise ValidationError("Invalid cuts = {}, not all members in baks"
                                  " for evt = {}.".format(cuts, ked))

        adds = ked["ba"]
        addset = oset(adds)
        if len(addset) != len(adds):
            raise ValidationError("Invalid adds = {}, has duplicates for evt = "
                                  "{}.".format(adds, ked))

        if cutset & addset:  # non empty intersection
            raise ValidationError("Intersecting cuts = {} and  adds = {} for "
                                  "evt = {}.".format(cuts, adds, ked))

        if witset & addset:  # non empty intersection
            raise ValidationError("Intersecting baks = {} and  adds = {} for "
                                  "evt = {}.".format(self.baks, adds, ked))

        baks = list((witset - cutset) | addset)

        if len(baks) != (len(self.baks) - len(cuts) + len(adds)):  # redundant?
            raise ValidationError("Invalid member combination among baks = {}, cuts ={}, "
                                  "and adds = {} for evt = {}.".format(self.baks,
                                                                       cuts,
                                                                       adds,
                                                                       ked))

        toad = int(ked["bt"], 16)
        if baks:
            if toad < 1 or toad > len(baks):  # out of bounds toad
                raise ValidationError("Invalid toad = {} for baks = {} for evt "
                                      "= {}.".format(toad, baks, ked))
        else:
            if toad != 0:  # invalid toad
                raise ValidationError("Invalid toad = {} for baks = {} for evt "
                                      "= {}.".format(toad, baks, ked))

        return toad, baks, cuts, adds

    def issue(self, serder, seqner, saider, sn, bigers=None):
        """ Process VC TEL issuance events (iss, bis)

        Validate and process credential issuance events.  If valid, event is persisted
        in local datastore for TEL.  Will escrow event if missing anchor or backer signatures

        Parameters
            serder (Serder): instance of issuance or backer issuance event
            seqner (Seqner): issuing event sequence number from controlling KEL.
            saider (Saider): issuing event SAID from controlling KEL.
            sn (int): event sequence event
            bigers (list): of Siger instances of indexed witness signatures.
                Index is offset into wits list of associated witness nontrans pre
                from which public key may be derived.

        """

        ked = serder.ked
        vcpre = ked["i"]
        ilk = ked["t"]
        vci = vcpre

        #labels = ISS_LABELS if ilk == Ilks.iss else BIS_LABELS
        #for k in labels:
            #if k not in ked:
                #raise ValidationError("Missing element = {} from {} event for "
                                      #"evt = {}.".format(k, ilk, ked))

        if ilk == Ilks.iss:  # simple issue
            if self.noBackers is False:
                raise ValidationError("invalid simple issue evt {} against backer based registry {}".
                                      format(ked, self.regk))

            regi = ked["ri"]
            if regi != self.prefixer.qb64:
                raise ValidationError("Mismatch event regi prefix = {} expecting"
                                      " = {} for evt = {}.".format(regi,
                                                                   self.prefixer.qb64,
                                                                   ked))

            # check if fully anchored
            if not self.verifyAnchor(serder=serder, seqner=seqner, saider=saider):
                self.escrowALEvent(serder=serder, seqner=seqner, saider=saider)
                raise MissingAnchorError("Failure verify event = {} "
                                         "".format(serder.ked,
                                                   ))

            self.logEvent(pre=vci, sn=sn, serder=serder, seqner=seqner, saider=saider)

        elif ilk == Ilks.bis:  # backer issue
            if self.noBackers is True:
                raise ValidationError("invalid backer issue evt {} against backerless registry {}".
                                      format(ked, self.regk))

            rtoad, baks = self.getBackerState(ked)
            bigers = self.valAnchorBigs(serder=serder,
                                        seqner=seqner,
                                        saider=saider,
                                        bigers=bigers,
                                        toad=rtoad,
                                        baks=baks)

            self.logEvent(pre=vci, sn=sn, serder=serder, seqner=seqner, saider=saider, bigers=bigers)

        else:
            raise ValidationError("Unsupported ilk = {} for evt = {}.".format(ilk, ked))

    def revoke(self, serder, seqner, saider, sn, bigers=None):
        """ Process VC TEL revocation events (rev, brv)

        Validate and process credential revocation events.  If valid, event is persisted
        in local datastore for TEL.  Will escrow event if missing anchor or backer signatures

        Parameters
            serder (Serder): instance of issuance or backer issuance event
            seqner (Seqner): issuing event sequence number from controlling KEL.
            saider (Saider): issuing event digest from controlling KEL.
            sn (int): event sequence event
            bigers (list): of Siger instances of indexed witness signatures.
                Index is offset into wits list of associated witness nontrans pre
                from which public key may be derived.

        """

        ked = serder.ked
        vcpre = ked["i"]
        ilk = ked["t"]

        #labels = REV_LABELS if ilk == Ilks.rev else BRV_LABELS
        #for k in labels:
            #if k not in ked:
                #raise ValidationError("Missing element = {} from {} event for "
                                      #"evt = {}.".format(k, ilk, ked))

        # have to compare with VC issuance serder
        vci = vcpre

        dig = self.reger.getTel(snKey(pre=vci, sn=sn - 1))
        ievt = self.reger.getTvt(dgKey(pre=vci, dig=dig))
        if ievt is None:
            raise ValidationError("revoke without issue... probably have to escrow")

        ievt = bytes(ievt)
        iserder = serdering.SerderKERI(raw=ievt)
        if not iserder.compare(said=ked["p"]):  # prior event dig not match
            raise ValidationError("Mismatch event dig = {} with state dig"
                                  " = {} for evt = {}.".format(ked["p"],
                                                               self.serder.said,
                                                               ked))

        if ilk in (Ilks.rev,):  # simple revoke
            if self.noBackers is False:
                raise ValidationError("invalid simple issue evt {} against backer based registry {}".
                                      format(ked, self.regk))

            # check if fully anchored
            if not self.verifyAnchor(serder=serder, seqner=seqner, saider=saider):
                self.escrowALEvent(serder=serder, seqner=seqner, saider=saider)
                raise MissingAnchorError("Failure verify event = {} "
                                         "".format(serder.ked))

            self.logEvent(pre=vci, sn=sn, serder=serder, seqner=seqner, saider=saider)
            self.cues.push(dict(kin="revoked", serder=serder))

        elif ilk in (Ilks.brv,):  # backer revoke
            if self.noBackers is True:
                raise ValidationError("invalid backer issue evt {} against backerless registry {}".
                                      format(ked, self.regk))

            rtoad, baks = self.getBackerState(ked)
            bigers = self.valAnchorBigs(serder=serder,
                                        seqner=seqner,
                                        saider=saider,
                                        bigers=bigers,
                                        toad=rtoad,
                                        baks=baks)

            self.logEvent(pre=vci, sn=sn, serder=serder, seqner=seqner, saider=saider, bigers=bigers)
            self.cues.push(dict(kin="revoked", serder=serder))

        else:
            raise ValidationError("Unsupported ilk = {} for evt = {}.".format(ilk, ked))

    def vcState(self, vci):
        """ Calculate state (issued/revoked) of VC from db.

        Returns None if never issued from this Registry

        Parameters:
          vci (str):  qb64 VC identifier

        Returns:
            status (Serder): transaction event state notification message
        """
        digs = []
        for _, _, dig in self.reger.getTelItemPreIter(pre=vci.encode("utf-8")):
            digs.append(dig)

        if len(digs) == 0:
            return None

        vcsn = len(digs) - 1
        vcdig = bytes(digs[-1])

        dgkey = dbing.dgKey(vci, vcdig)  # get message
        raw = self.reger.getTvt(key=dgkey)
        serder = serdering.SerderKERI(raw=bytes(raw))

        if self.noBackers:
            vcilk = Ilks.iss if len(digs) == 1 else Ilks.rev
            ra = dict()
        else:
            vcilk = Ilks.bis if len(digs) == 1 else Ilks.brv
            ra = serder.ked["ra"]

        dgkey = dbing.dgKey(vci, vcdig)
        couple = self.reger.getAnc(dgkey)
        ancb = bytearray(couple)
        seqner = coring.Seqner(qb64b=ancb, strip=True)
        saider = coring.Saider(qb64b=ancb, strip=True)

        return vcstate(vcpre=vci,
                       said=vcdig.decode("utf-8"),
                       sn=vcsn,
                       ri=self.prefixer.qb64,
                       dts=serder.ked['dt'],
                       eilk=vcilk,
                       ra=ra,
                       a=dict(s=seqner.sn, d=saider.qb64),
                       )

    def vcSn(self, vci):
        """ Calculates the current seq no of VC from db.

        Returns None if never issued from this Registry

        Parameters:
          vci (str):  qb64 VC identifier

        Returns:
            int: current TEL sequence number of credential or None if not found

        """
        cnt = self.reger.cntTels(vci)

        return None if cnt == 0 else cnt - 1

    def logEvent(self, pre, sn, serder, seqner, saider, bigers=None, baks=None):
        """ Update associated logs for verified event.

        Update is idempotent. Logs will not write dup at key if already exists.

        Parameters:
            pre (str): is event prefix
            sn (int): is event sequence number
            serder (Serder): is Serder instance of current event
            seqner (Seqner): issuing event sequence number from controlling KEL.
            saider (Saider): issuing event SAID from controlling KEL.
            bigers (list): is optional list of Siger instance of indexed backer sigs
            baks (list): is optional list of qb64 non-trans identifiers of backers
        """
        if hasattr(pre, "encode"):
            pre = pre.encode("utf-8")  # convert str to bytes

        dig = serder.saidb
        key = dgKey(pre, dig)
        sealet = seqner.qb64b + saider.qb64b
        self.reger.putAnc(key, sealet)
        if bigers:
            self.reger.putTibs(key, [biger.qb64b for biger in bigers])
        if baks:
            self.reger.delBaks(key)
            self.reger.putBaks(key, [bak.encode("utf-8") for bak in baks])
        self.reger.tets.pin(keys=(pre.decode("utf-8"), dig.decode("utf-8")), val=coring.Dater())
        self.reger.putTvt(key, serder.raw)
        self.reger.putTel(snKey(pre, sn), dig)
        logger.info("Tever state: %s Added to TEL valid said=%s",
                    pre, serder.said)
        logger.debug(f"event=\n{serder.pretty()}\n")

    def valAnchorBigs(self, serder, seqner, saider, bigers, toad, baks):
        """ Validate anchor and backer signatures (bigers) when provided.

        Validates sigers signatures by validating indexes, verifying signatures, and
            validating threshold sith.
        Validate backer receipts by validating indexes, verifying
            backer signatures and validating toad.
        Backer validation is a function of .regk and .local

        Parameters:
            serder (Serder): instance of event
            seqner (Seqner): issuing event sequence number from controlling KEL.
            saider (Saider): issuing event said from controlling KEL.
            bigers (list)  Siger instances of indexed witness signatures.
                Index is offset into wits list of associated witness nontrans pre
                from which public key may be derived.
            toad (int):  str hex of witness threshold
            baks (list): qb64 non-transferable prefixes of backers used to
                derive werfers for bigers

        Returns:
            list: unique validated signature verified members of inputed bigers

        """

        berfers = [Verfer(qb64=bak) for bak in baks]

        # get unique verified bigers and bindices lists from bigers list
        bigers, bindices = verifySigs(raw=serder.raw, sigers=bigers, verfers=berfers)
        # each biger now has werfer of corresponding wit

        # check if fully anchored
        if not self.verifyAnchor(serder=serder, seqner=seqner, saider=saider):
            self.escrowALEvent(serder=serder, seqner=seqner, saider=saider, bigers=bigers, baks=baks)
            raise MissingAnchorError("Failure verify event = {}".format(serder.ked))

        # Kevery .process event logic prevents this from seeing event when
        # not local and event pre is own pre
        if ((baks and not self.regk) or  # in promiscuous mode so assume must verify toad
                (baks and not self.local and self.regk and self.regk not in baks)):
            # validate that event is fully witnessed
            if isinstance(toad, str):
                toad = int(toad, 16)
            if toad < 0 or len(baks) < toad:
                raise ValidationError("Invalid toad = {} for wits = {} for evt"
                                      " = {}.".format(toad, baks, serder.ked))

            if len(bindices) < toad:  # not fully witnessed yet
                self.escrowPWEvent(serder=serder, seqner=seqner, saider=saider, bigers=bigers)

                raise MissingWitnessSignatureError("Failure satisfying toad = {} "
                                                   "on witness sigs for {} for evt = {}.".format(toad,
                                                                                                 [siger.qb64 for siger
                                                                                                  in bigers],
                                                                                                 serder.ked))
        return bigers

    def verifyAnchor(self, serder, seqner=None, saider=None):
        """ Retrieve specified anchoring event and verify seal

        Retrieve event from db using anchor, get seal from event eserder and
        verify pre, sn and dig against serder

        Parameters:
            serder (Serder): anchored TEL event
            seqner (Seqner): sequence number of anchoring event
            saider (Saider): digest of anchoring event

        Returns:
             bool: True is anchoring event exists in database and seal is valid against
                   TEL event.

        """

        if seqner is None or saider is None:
            return False

        dig = self.db.getKeLast(key=snKey(pre=self.pre, sn=seqner.sn))
        if not dig:
            return False
        else:
            dig = bytes(dig)

        # retrieve event by dig
        raw = self.db.getEvt(key=dgKey(pre=self.pre, dig=dig))
        if not raw:
            return False
        else:
            raw = bytes(raw)

        eserder = serdering.SerderKERI(raw=raw)  # deserialize event raw

        if eserder.said != saider.qb64:
            return False

        seal = eserder.ked["a"]
        if seal is None or len(seal) != 1:
            return False

        seal = seal[0]
        spre = seal["i"]
        ssn = seal["s"]
        sdig = seal["d"]

        if spre == serder.ked["i"] and ssn == serder.ked["s"] \
                and serder.said == sdig:
            return True

        return False

    def escrowPWEvent(self, serder, seqner, saider, bigers=None):
        """ Update associated logs for escrow of partially witnessed event

        Parameters:
            serder (Serder): instance of  event
            seqner (Seqner): sequence number for anchor seal
            saider (Saider): digest of anchor
            bigers (list): Siger instance of indexed witness sigs

        """
        dgkey = dgKey(serder.preb, serder.saidb)
        sealet = seqner.qb64b + saider.qb64b
        self.reger.putAnc(dgkey, sealet)
        self.reger.putTibs(dgkey, [biger.qb64b for biger in bigers])
        self.reger.putTvt(dgkey, serder.raw)
        self.reger.putTwe(snKey(serder.preb, serder.sn), serder.saidb)
        logger.debug("Tever state: Escrowed partially witnessed "
                     "event = %s", serder.ked)

    def escrowALEvent(self, serder, seqner, saider, bigers=None, baks=None):
        """ Update associated logs for escrow of anchorless event

        Parameters:
            serder (Serder): instance of  event
            seqner (Seqner): sequence number for anchor seal
            saider (Saider): SAID of anchor
            bigers (list): Siger instance of indexed witness sigs
            baks (list): qb64 of new backers

        Returns:
            bool: True if escrow is successful, False otherwith (eg. already escrowed)

        """
        key = dgKey(serder.preb, serder.saidb)
        if seqner and saider:
            sealet = seqner.qb64b + saider.qb64b
            self.reger.putAnc(key, sealet)
        if bigers:
            self.reger.putTibs(key, [biger.qb64b for biger in bigers])
        if baks:
            self.reger.delBaks(key)
            self.reger.putBaks(key, [bak.encode("utf-8") for bak in baks])
        self.reger.putTvt(key, serder.raw)
        logger.debug("Tever state: Escrowed anchorless event "
                     "event = %s", serder.ked)
        return self.reger.putTae(snKey(serder.preb, serder.sn), serder.saidb)

    def getBackerState(self, ked):
        """ Calculate and return the current list of backers for event dict

        Parameters:
            ked (dict):  event dict

        Returns:
            list:  qb64 of current list of backers for state at ked

        """
        rega = ked["ra"]
        regi = rega["i"]
        regd = rega["d"]

        if regi != self.prefixer.qb64:
            raise ValidationError("Mismatch event regk prefix = {} expecting"
                                  " = {} for evt = {}.".format(self.regk,
                                                               self.prefixer.qb64,
                                                               ked))

        # load backer list and toad (via event) for specific event in registry from seal in event
        dgkey = dgKey(regi, regd)
        revt = self.reger.getTvt(dgkey)
        if revt is None:
            raise ValidationError("have to escrow this somewhere")

        rserder = serdering.SerderKERI(raw=bytes(revt))
        # the backer threshold at this event in mgmt TEL
        rtoad = rserder.ked["bt"]

        baks = [bytes(bak).decode("utf-8") for bak in self.reger.getBaks(dgkey)]

        return rtoad, baks


class Tevery:
    """ Tevery (Transaction Event Message Processing Facility)

    Tevery processes an incoming message stream composed of KERI key event related
    messages and attachments.  Tevery acts as a Tever (transaction event verifier)
    factory for managing transaction state of KERI credential registries and associated
    credentials.

    Attributes:
        db (Baser):  local LMDB identifier database
        reger (Reger): local LMDB credential database
        local (bool): True means only process msgs for own events if .regk
                        False means only process msgs for not own events if .regk
        cues (Deck): notices generated from processing events


    """

    TimeoutTSN = 3600

    def __init__(self, reger=None, db=None, local=False, lax=False, cues=None, rvy=None):
        """ Initialize instance:

        Parameters:
            reger (Reger): local LMDB credential database
            db (Baser):  local LMDB identifier database
            local (bool): True means only process msgs for own events if .regk
                        False means only process msgs for not own events if .regk
            cues (Deck): notices generated from processing events


        """
        self.db = db if db is not None else basing.Baser(reopen=True)  # default name = "main"
        self.rvy = rvy
        self.reger = reger if reger is not None else viring.Reger()
        self.local = True if local else False  # local vs nonlocal restrictions
        self.lax = True if lax else False
        self.cues = cues if cues is not None else decking.Deck()

    @property
    def tevers(self):
        """ Returns .reger.tevers read through cache of credential registries """

        return self.reger.tevers

    @property
    def kevers(self):
        """ Returns .db.kevers read through cache of key event logs """

        return self.db.kevers

    @property
    def registries(self):
        """ Returns .reger.registries """

        return self.reger.registries

    def processEvent(self, serder, seqner=None, saider=None, wigers=None):
        """ Process one event serder with attached indexed signatures sigers

        Validates event against current state of registry or credential, creating registry
        on inception events and processing change in state to credential or registry for
        other events

        Parameters:
            serder (Serder): event to process
            seqner (Seqner): issuing event sequence number from controlling KEL.
            saider (Saider): issuing event digest from controlling KEL.
            wigers (list): optional list of Siger instances of attached witness indexed sigs

        """
        ked = serder.ked
        try:  # see if code of pre is supported and matches size of pre
            Prefixer(qb64b=serder.preb)
        except Exception:  # if unsupported code or bad size raises error
            raise ValidationError("Invalid pre = {} for evt = {}."
                                  "".format(serder.pre, ked))

        regk = self.registryKey(serder)
        pre = serder.pre
        ked = serder.ked
        #sn = ked["s"]
        ilk = ked["t"]

        inceptive = ilk in (Ilks.vcp, Ilks.iss, Ilks.bis)

        # validate SN for
        #sn = validateSN(sn, inceptive=inceptive)
        sn = Number(numh=ked["s"]).validate(inceptive=inceptive).sn

        if not self.lax:
            if self.local:
                if regk not in self.registries:  # nonlocal event when in local mode
                    raise ValueError("Nonlocal event regk={} when local mode for registries={}."
                                     "".format(regk, self.registries))
            else:
                if regk in self.registries:  # local event when not in local mode
                    raise ValueError("Local event regk={} when nonlocal mode."
                                     "".format(regk))

        if regk not in self.tevers:  # first seen for this registry
            if ilk in [Ilks.vcp]:
                # incepting a new registry, Tever create will validate anchor, etc.
                tever = Tever(serder=serder,
                              seqner=seqner,
                              saider=saider,
                              bigers=wigers,
                              reger=self.reger,
                              db=self.db,
                              regk=regk,
                              local=self.local,
                              cues=self.cues)
                self.tevers[regk] = tever
                if regk not in self.registries:
                    # witness style backers will need to send receipts so lets queue them up for now
                    # actually, lets not because the Kevery has no idea what to do with them!
                    # self.cues.append(dict(kin="receipt", serder=serder))
                    pass
            else:
                # out of order, need to escrow
                self.escrowOOEvent(serder=serder, seqner=seqner, saider=saider)
                raise OutOfOrderError("escrowed out of order event {}".format(ked))

        else:
            if ilk in (Ilks.vcp,):
                # we don't have multiple signatures to verify so this
                # is already first seen and then lifely duplicitious
                raise LikelyDuplicitousError("Likely Duplicitous event={}.".format(ked))

            tever = self.tevers[regk]
            tever.cues = self.cues
            if ilk in [Ilks.vrt]:
                sno = tever.sn + 1  # proper sn of new inorder event
            else:
                esn = tever.vcSn(pre)
                sno = 0 if esn is None else esn + 1

            #if not serder.saider.verify(sad=serder.sad):
                #raise ValidationError("Invalid SAID {} for event {}".format(said, serder.ked))

            if sn > sno:  # sn later than sno so out of order escrow
                # escrow out-of-order event
                self.escrowOOEvent(serder=serder, seqner=seqner, saider=saider)
                raise OutOfOrderError("Out-of-order event={}.".format(ked))
            elif sn == sno:  # new inorder event
                tever.update(serder=serder, seqner=seqner, saider=saider, bigers=wigers)

                if regk not in self.registries:
                    # witness style backers will need to send receipts so lets queue them up for now
                    # actually, lets not because the Kevery has no idea what to do with them!
                    # self.cues.append(dict(kin="receipt", serder=serder))
                    pass
            else:  # duplicitious
                raise LikelyDuplicitousError("Likely Duplicitous event={} with sn {}.".format(ked, sn))

    def processQuery(self, serder, source=None, sigers=None, cigars=None):
        """ Process TEL query event message (qry)

        Process query mode replay message for collective or single element query.
        Will cue response message with kin of "replay".  Assume promiscuous mode for now.

        Parameters:
            serder (Serder): is query message serder
            source (qb64): identifier prefix of querier
            sigers (list): Siger instances of attached controller indexed sigs
            cigars (list): Siger instances of non-transferable signatures

        """
        ked = serder.ked

        ilk = ked["t"]
        route = ked["r"]
        qry = ked["q"]

        # do signature validation and replay attack prevention logic here
        # src, dt, route

        if route == "tels":
            mgmt = qry["ri"]
            src = qry["src"]

            cloner = self.reger.clonePreIter(pre=mgmt, fn=0)  # create iterator at 0
            msgs = list()  # outgoing messages
            for msg in cloner:
                msgs.append(msg)

            if vci := qry["i"]:
                cloner = self.reger.clonePreIter(pre=vci, fn=0)  # create iterator at 0
                for msg in cloner:
                    msgs.append(msg)

            if msgs:
                self.cues.append(dict(kin="replay", src=src, dest=source.qb64, msgs=msgs))
        elif route == "tsn":
            ri = qry["ri"]
            if ri in self.tevers:
                tever = self.tevers[ri]
                tsn = tever.state()
                self.cues.push(dict(kin="reply", route="/tsn/registry", data=asdict(tsn), dest=source))

                if vcpre := qry["i"]:
                    tsn = tever.vcState(vcpre=vcpre)
                    self.cues.push(dict(kin="reply", route="/tsn/credential", data=asdict(tsn), dest=source))

        else:
            raise ValidationError("invalid query message {} for evt = {}".format(ilk, ked))

    def registerReplyRoutes(self, router):
        """ Register the routes for processing messages embedded in `rpy` event messages

        Parameters:
            router(Router): reply message router

        """
        router.addRoute("/tsn/registry/{aid}", self, suffix="RegistryTxnState")
        router.addRoute("/tsn/credential/{aid}", self, suffix="CredentialTxnState")

    def processReplyRegistryTxnState(self, *, serder, saider, route, cigars=None, tsgs=None, **kwargs):
        """ Process one reply message for key state = /tsn/registry

         Process one reply message for key state = /tsn/registry
         with either attached nontrans receipt couples in cigars or attached trans
         indexed sig groups in tsgs.
         Assumes already validated saider, dater, and route from serder.ked

         Parameters:
             serder (Serder): instance of reply msg (SAD)
             saider (Saider): instance  from said in serder (SAD)
             route (str): reply route
             cigars (list): of Cigar instances that contain nontrans signing couple
                           signature in .raw and public key in .verfer
             tsgs (list): tuples (quadruples) of form
                 (prefixer, seqner, diger, [sigers]) where:
                 prefixer is pre of trans endorser
                 seqner is sequence number of trans endorser's est evt for keys for sigs
                 diger is digest of trans endorser's est evt for keys for sigs
                 [sigers] is list of indexed sigs from trans endorser's keys from est evt

         Reply Message:
         {
           "v" : "KERI10JSON00011c_",
           "t" : "rpy",
           "d": "EZ-i0d8JZAoTNZH3ULaU6JR2nmwyvYAfSVPzhzS6b5CM",
           "dt": "2020-08-22T17:50:12.988921+00:00",
           "r" : "/tsn/EgHOJJ9mgNosU2hgt7bsM8AViwgz--ey3ZXWgfIcxdpI",
           "a" :
             {
               "v": "KERI10JSON0001b0_",
               "i": "EoN_Ln_JpgqsIys-jDOH8oWdxgWqs7hzkDGeLWHb9vSY",
               "s": "1",
               "d": "EpltHxeKueSR1a7e0_oSAhgO6U7VDnX7x4KqNCwBqbI0",
               "ii": "EaKJ0FoLxO1TYmyuprguKO7kJ7Hbn0m0Wuk5aMtSrMtY",
               "dt": "2021-01-01T00:00:00.000000+00:00",
               "et": "vrt",
               "a": {
                "s": 2,
                "d": "Ef12IRHtb_gVo5ClaHHNV90b43adA0f8vRs3jeU-AstY"
               },
               "bt": "1",
               "br": [],
               "ba": [
                "BwFbQvUaS4EirvZVPUav7R_KDHB8AKmSfXNpWnZU_YEU"
               ],
               "b": [
                "BwFbQvUaS4EirvZVPUav7R_KDHB8AKmSfXNpWnZU_YEU"
               ],
               "c": []
             }
         }

         """
        cigars = cigars if cigars is not None else []
        tsgs = tsgs if tsgs is not None else []

        # reply specific logic
        if not route.startswith("/tsn"):
            raise ValidationError(f"Usupported route={route} in {Ilks.rpy} "
                                  f"msg={serder.ked}.")
        aid = kwargs["aid"]
        data = serder.ked["a"]
        dater = coring.Dater(dts=serder.ked["dt"])

        rsr = viring.RegStateRecord(**data)

        # fetch from serder to process
        regk = rsr.i
        pre = rsr.ii
        sn = int(rsr.s, 16)

        if pre not in self.kevers:
            if self.reger.txnsb.escrowStateNotice(typ="registry-mae", pre=regk, aid=aid, serder=serder, saider=saider,
                                                  dater=dater, cigars=cigars, tsgs=tsgs):
                self.cues.append(dict(kin="query", q=dict(pre=pre)))

            raise kering.MissingAnchorError("Failure verify event = {} ".format(serder.ked))

        # Load backers from either tsn or Kever of issuer
        cnfg = rsr.c
        if TraitDex.NoBackers in cnfg:
            kevers = self.kevers[pre]
            baks = kevers.wits
        else:
            baks = rsr.b

        wats = set()
        for _, habr in self.db.habs.getItemIter():
            wats |= set(habr.watchers)

        # not in promiscuous mode
        if not self.lax:
            # check source and ensure we should accept it
            if aid != pre and \
                    aid not in baks and \
                    aid not in wats:
                raise kering.UntrustedKeyStateSource("transaction state notice for {} from untrusted source {} "
                                                     .format(rsr.i, aid))

        if regk in self.tevers:
            tever = self.tevers[regk]
            if int(rsr.s, 16) < tever.sn:
                raise ValidationError("Skipped stale transaction state at sn {} for {}."
                                      "".format(rsr.s, rsr.i))

        keys = (regk, aid,)
        osaider = self.reger.txnsb.current(keys=keys)  # get old said if any

        # BADA Logic
        accepted = self.rvy.acceptReply(serder=serder, saider=saider, route=route,
                                        aid=aid, osaider=osaider, cigars=cigars,
                                        tsgs=tsgs)
        if not accepted:
            raise kering.UnverifiedReplyError(f"Unverified registry txn state reply.")

        ldig = self.reger.getTel(key=snKey(pre=regk, sn=sn))  # retrieve dig of last event at sn.

        # Only accept key state if for last seen version of event at sn
        if ldig is None:  # escrow because event does not yet exist in database
            if self.reger.txnsb.escrowStateNotice(typ="registry-ooo", pre=regk, aid=aid, serder=serder, saider=saider,
                                                  dater=dater, cigars=cigars, tsgs=tsgs):
                self.cues.append(dict(kin="telquery", q=dict(ri=regk)))

            raise kering.OutOfOrderTxnStateError("Out of order txn state={}.".format(rsr))

        tsaider = coring.Saider(qb64=rsr.d)
        ldig = bytes(ldig)
        # retrieve last event itself of signer given sdig
        sraw = self.reger.getTvt(key=dgKey(pre=regk, dig=ldig))
        # assumes db ensures that sraw must not be none because sdig was in KE
        sserder = serdering.SerderKERI(raw=bytes(sraw))

        if sserder.said != tsaider.qb64:  # mismatch events problem with replay
            raise ValidationError("Mismatch keystate at sn = {} with db."
                                  "".format(rsr.s))

        self.reger.txnsb.updateReply(aid=aid, serder=serder, saider=tsaider, dater=dater)
        self.cues.append(dict(kin="txnStateSaved", record=rsr))

    def processReplyCredentialTxnState(self, *, serder, saider, route, cigars=None, tsgs=None, **kwargs):
        """ Process one reply message for key state = /tsn/registry

         Process one reply message for key state = /tsn/registry
         with either attached nontrans receipt couples in cigars or attached trans
         indexed sig groups in tsgs.
         Assumes already validated saider, dater, and route from serder.ked

         Parameters:
             serder (Serder): instance of reply msg (SAD)
             saider (Saider): instance  from said in serder (SAD)
             route (str): reply route
             cigars (list): of Cigar instances that contain nontrans signing couple
                           signature in .raw and public key in .verfer
             tsgs (list): tuples (quadruples) of form
                 (prefixer, seqner, diger, [sigers]) where:
                 prefixer is pre of trans endorser
                 seqner is sequence number of trans endorser's est evt for keys for sigs
                 diger is digest of trans endorser's est evt for keys for sigs
                 [sigers] is list of indexed sigs from trans endorser's keys from est evt

         Reply Message:
         {
           "v" : "KERI10JSON00011c_",
           "t" : "rpy",
           "d": "EZ-i0d8JZAoTNZH3ULaU6JR2nmwyvYAfSVPzhzS6b5CM",
           "dt": "2020-08-22T17:50:12.988921+00:00",
           "r" : "/tsn/EgHOJJ9mgNosU2hgt7bsM8AViwgz--ey3ZXWgfIcxdpI",
           "a" :
             {
              "v": "KERI10JSON00012d_",
              "i": "EDGhJ8V1tuwH55Bk0fBFe9L0za2BUNOt2FX4GUeOLNHQ",
              "s": "0",
              "d": "ENNTabgWbaNqOKLqEZdQCjxbafwwSoXNzAsE1Enq-kdk",
              "ri": "EoN_Ln_JpgqsIys-jDOH8oWdxgWqs7hzkDGeLWHb9vSY",
              "a": {
               "s": 3,
               "d": "Ex7i6wv4YzDRTO9_iHkTQSXrvLYldSd_UEjNfqia3Pqc"
              },
              "dt": "2021-01-01T00:00:00.000000+00:00",
              "et": "bis"
             }
         }

         """
        cigars = cigars if cigars is not None else []
        tsgs = tsgs if tsgs is not None else []

        # reply specific logic
        if not route.startswith("/tsn"):
            raise ValidationError(f"Usupported route={route} in {Ilks.rpy} "
                                  f"msg={serder.ked}.")
        aid = kwargs["aid"]
        data = serder.ked["a"]
        dater = coring.Dater(dts=serder.ked["dt"])

        vsr = viring.VcStateRecord(**data)

        # fetch from serder to process
        regk = vsr.ri
        vci = vsr.i
        sn = int(vsr.s, 16)
        ra = vsr.ra

        if 's' in ra:
            regsn = int(ra["s"], 16)
        else:
            regsn = 0

        if regk not in self.tevers or self.tevers[regk].sn < regsn:
            if self.reger.txnsb.escrowStateNotice(typ="credential-mre", pre=vci, aid=aid, serder=serder,
                                                  saider=saider, dater=dater, cigars=cigars, tsgs=tsgs):
                self.cues.append(dict(kin="telquery", q=dict(ri=regk)))

            raise kering.MissingRegistryError("Failure verify event = {} ".format(serder.ked))

        tever = self.tevers[regk]
        pre = tever.pre

        if pre not in self.kevers:
            if self.reger.txnsb.escrowStateNotice(typ="credential-mae", pre=vci, aid=aid, serder=serder,
                                                  saider=saider, dater=dater, cigars=cigars, tsgs=tsgs):
                self.cues.append(dict(kin="query", q=dict(pre=aid)))

            raise kering.MissingAnchorError("Failure verify event = {} ".format(serder.ked))

        # Load backers from either tsn or Kever of issuer
        if tever.noBackers:
            kevers = self.kevers[pre]
            baks = kevers.wits
        else:
            baks = tever.baks

        wats = set()
        for _, habr in self.db.habs.getItemIter():
            wats |= set(habr.watchers)

        # not in promiscuous mode
        if not self.lax:
            # check source and ensure we should accept it
            if aid != pre and \
                    aid not in baks and \
                    aid not in wats:
                raise kering.UntrustedKeyStateSource("transaction state notice for {} from untrusted source {} "
                                                     .format(vsr.i, aid))

        keys = (vci, aid,)
        osaider = self.reger.txnsb.current(keys=keys)  # get old said if any

        # BADA Logic
        accepted = self.rvy.acceptReply(serder=serder, saider=saider, route=route,
                                        aid=aid, osaider=osaider, cigars=cigars,
                                        tsgs=tsgs)
        if not accepted:
            raise kering.UnverifiedReplyError(f"Unverified credential state reply.")

        ldig = self.reger.getTel(key=snKey(pre=vci, sn=sn))  # retrieve dig of last event at sn.

        # Only accept key state if for last seen version of event at sn
        if ldig is None:  # escrow because event does not yet exist in database
            if self.reger.txnsb.escrowStateNotice(typ="credential-ooo", pre=vci, aid=aid, serder=serder,
                                                  saider=saider, dater=dater, cigars=cigars, tsgs=tsgs):
                self.cues.append(dict(kin="telquery", q=dict(ri=regk, i=vci)))

            raise kering.OutOfOrderTxnStateError("Out of order txn state={}.".format(vsr))

        tsaider = coring.Saider(qb64=vsr.d)
        ldig = bytes(ldig)
        # retrieve last event itself of signer given sdig
        sraw = self.reger.getTvt(key=dgKey(pre=vci, dig=ldig))
        # assumes db ensures that sraw must not be none because sdig was in KE
        sserder = serdering.SerderKERI(raw=bytes(sraw))

        if sn < sserder.sn:
            raise ValidationError("Stale txn state at sn = {} with db."
                                  "".format(vsr.s))

        if sserder.said != tsaider.qb64:  # mismatch events problem with replay
            raise ValidationError("Mismatch txn state at sn = {} with db."
                                  "".format(vsr.s))

        self.reger.txnsb.updateReply(aid=aid, serder=serder, saider=tsaider, dater=dater)
        self.cues.append(dict(kin="txnStateSaved", record=vsr))

    @staticmethod
    def registryKey(serder):
        """  Utility method to extract registry key from any type of TEL serder

        Parameters:
            serder (Serder): event messate

        Returns:
            str: qb64 registry identifier
        """
        ilk = serder.ked["t"]

        if ilk in (Ilks.vcp, Ilks.vrt):
            return serder.pre
        elif ilk in (Ilks.iss, Ilks.rev):
            return serder.ked["ri"]
        elif ilk in (Ilks.bis, Ilks.brv):
            rega = serder.ked["ra"]
            return rega["i"]
        else:
            raise ValidationError("invalid ilk {} for tevery event = {}".format(ilk, serder.ked))

    def escrowOOEvent(self, serder, seqner, saider):
        """ Escrow out-of-order TEL events.

        Saves the serialized event, anchor and event digest in escrow for any
        event that is received out of order.

        Examples include registry rotation events, credential issuance event
         received before the registry inception event or a credential revocation
         event received before the issuance event.

        Parameters:
            serder (Serder): serder of event message
            seqner (Seqner): sequence number of anchoring TEL event
            saider (Diger) digest of anchoring TEL event


        """
        key = dgKey(serder.preb, serder.saidb)
        self.reger.putTvt(key, serder.raw)
        sealet = seqner.qb64b + saider.qb64b
        self.reger.putAnc(key, sealet)
        self.reger.putOot(snKey(serder.preb, serder.sn), serder.saidb)
        logger.debug("Tever state: Escrowed our of order TEL event "
                     "event = %s", serder.ked)

    def processEscrows(self):
        """ Loop through escrows and process and events that may now be finalized """

        try:
            self.processEscrowAnchorless()
            self.processEscrowOutOfOrders()
            self.reger.txnsb.processEscrowState(typ="credential-mre", processReply=self.processReplyCredentialTxnState,
                                                extype=kering.MissingRegistryError)
            self.reger.txnsb.processEscrowState(typ="credential-mae", processReply=self.processReplyCredentialTxnState,
                                                extype=kering.MissingAnchorError)
            self.reger.txnsb.processEscrowState(typ="credential-ooo", processReply=self.processReplyCredentialTxnState,
                                                extype=kering.OutOfOrderTxnStateError)
            self.reger.txnsb.processEscrowState(typ="registry-mae", processReply=self.processReplyRegistryTxnState,
                                                extype=kering.MissingAnchorError)
            self.reger.txnsb.processEscrowState(typ="registry-ooo", processReply=self.processReplyRegistryTxnState,
                                                extype=kering.OutOfOrderTxnStateError)

        except Exception as ex:  # log diagnostics errors etc
            if logger.isEnabledFor(logging.DEBUG):
                logger.exception("Tevery escrow process error: %s", ex.args[0])
            else:
                logger.error("Tevery escrow process error: %s", ex.args[0])

    def processEscrowOutOfOrders(self):
        """ Loop through out of order escrow:

         Process out of order events in the following way:
           1. loop over event digests saved in oots
           2. deserialize event out of tvts
           3. read anchor information out of .ancs
           4. perform process event
           5. Remove event digest from oots if processed successfully or a non-out-of-order event occurs.

        """
        for key, digb in self.reger.getOotItemIter(): # (pre, snb, digb) in self.reger.getOotItemIter()
            try:
                #sn = int(snb, 16)
                pre, sn = splitSnKey(key)
                dgkey = dgKey(pre, digb)
                traw = self.reger.getTvt(dgkey)
                if traw is None:
                    # no event so raise ValidationError which unescrows below
                    logger.info("Tevery unescrow error: Missing event at."
                                "dig = %s", bytes(digb))

                    raise ValidationError("Missing escrowed evt at dig = {}."
                                          "".format(bytes(digb)))

                tserder = serdering.SerderKERI(raw=bytes(traw))  # escrowed event

                bigers = None
                if tibs := self.reger.getTibs(key=dgkey):
                    bigers = [indexing.Siger(qb64b=tib) for tib in tibs]

                couple = self.reger.getAnc(dgkey)
                if couple is None:
                    logger.info("Tevery unescrow error: Missing anchor at."
                                "dig = %s", bytes(digb))

                    raise ValidationError("Missing escrowed anchor at dig = {}."
                                          "".format(bytes(digb)))
                ancb = bytearray(couple)
                seqner = coring.Seqner(qb64b=ancb, strip=True)
                saider = coring.Saider(qb64b=ancb, strip=True)

                self.processEvent(serder=tserder, seqner=seqner, saider=saider, wigers=bigers)

            except OutOfOrderError as ex:
                # still waiting on missing prior event to validate
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception("Tevery unescrow failed: %s", ex.args[0])
                else:
                    logger.error("Tevery unescrow failed: %s", ex.args[0])

            except Exception as ex:  # log diagnostics errors etc
                # error other than out of order so remove from OO escrow
                self.reger.delOot(snKey(pre, sn))  # removes one escrow at key val
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception("Tevery unescrowed: %s", ex.args[0])
                else:
                    logger.error("Tevery unescrowed: %s", ex.args[0])

            else:  # unescrow succeeded, remove from escrow
                # We don't remove all escrows at pre,sn because some might be
                # duplicitous so we process remaining escrows in spite of found
                # valid event escrow.
                self.reger.delOot(snKey(pre, sn))  # removes from escrow
                logger.info("Tevery unescrow succeeded in valid event: "
                            "said=%s", tserder.said)
                logger.debug(f"event=\n{tserder.pretty()}\n")

    def processEscrowAnchorless(self):
        """ Process escrow of TEL events received before the anchoring KEL event.

        Process anchorless events in the following way:
           1. loop over event digests saved in taes
           2. deserialize event out of tvts
           3. load backer signatures out of tibs
           4. read anchor information out of ancs
           5. perform process event
           6. Remove event digest from oots if processed successfully or a non-anchorless event occurs.

        """
        for key, digb in self.reger.getTaeItemIter():  #(pre, snb, digb) in self.reger.getTaeItemIter()
            pre, sn = splitSnKey(key)
            #sn = int(snb, 16)
            try:
                dgkey = dgKey(pre, digb)
                traw = self.reger.getTvt(dgkey)
                if traw is None:
                    # no event so raise ValidationError which unescrows below
                    logger.info("Tevery unescrow error: Missing event at."
                                "dig = %s", bytes(digb))

                    raise ValidationError("Missing escrowed evt at dig = {}."
                                          "".format(bytes(digb)))

                tserder = serdering.SerderKERI(raw=bytes(traw))  # escrowed event

                bigers = None
                if tibs := self.reger.getTibs(key=dgkey):
                    bigers = [indexing.Siger(qb64b=tib) for tib in tibs]

                couple = self.reger.getAnc(dgkey)
                if couple is None:
                    logger.info("Tevery unescrow error: Missing anchor at."
                                "dig = %s", bytes(digb))

                    raise MissingAnchorError("Missing escrowed anchor at dig = {}."
                                             "".format(bytes(digb)))
                ancb = bytearray(couple)
                seqner = coring.Seqner(qb64b=ancb, strip=True)
                saider = coring.Saider(qb64b=ancb, strip=True)

                self.processEvent(serder=tserder, seqner=seqner, saider=saider, wigers=bigers)

            except MissingAnchorError as ex:
                # still waiting on missing prior event to validate
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception("Tevery unescrow failed: %s", ex.args[0])
                else:
                    logger.error("Tevery unescrow failed: %s", ex.args[0])

            except Exception as ex:  # log diagnostics errors etc
                # error other than out of order so remove from OO escrow
                self.reger.delTae(snKey(pre, sn))  # removes one escrow at key val
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception("Tevery unescrowed: %s", ex.args[0])
                else:
                    logger.error("Tevery unescrowed: %s", ex.args[0])

            else:  # unescrow succeeded, remove from escrow
                # We don't remove all escrows at pre,sn because some might be
                # duplicitous so we process remaining escrows in spite of found
                # valid event escrow.
                self.reger.delTae(snKey(pre, sn))  # removes from escrow
                logger.info("Tevery unescrow succeeded in valid event: "
                            "said=%s", tserder.said)
                logger.debug(f"event=\n{tserder.pretty()}\n")



# -*- encoding: utf-8 -*-
"""
keri.app.cli.commands module

"""
import argparse

from hio.base import doing

import keri
from keri.app.cli.common import existing

parser = argparse.ArgumentParser(description='Print version of KLI')
parser.set_defaults(handler=lambda args: handler(args))
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=False,
                    default=None)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--passcode', '-p', help='21 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran


def handler(args):
    kwa = dict(args=args)
    return [doing.doify(version, **kwa)]


def version(tymth, tock=0.0, **opts):
    """ Command line version handler
    """
    _ = (yield tock)

    args = opts["args"]
    name = args.name
    base = args.base
    bran = args.bran

    print(f"Library version: {keri.__version__}")

    if name is not None:
        with existing.existingHby(name=name, base=base, bran=bran) as hby:
            print(f"Database version: {hby.db.version}")



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""
import argparse

from hio import help
from hio.base import doing

from keri import kering
from keri.app import habbing, grouping, indirecting, forwarding
from keri.app.agenting import WitnessPublisher
from keri.app.cli.common import existing
from keri.app.notifying import Notifier
from keri.core import parsing
from keri.peer import exchanging

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Add new endpoint role authorization.')
parser.set_defaults(handler=lambda args: add_end(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--alias', '-a', help='human readable alias for the new identifier prefix', required=True)
parser.add_argument('--passcode', '-p', help='21 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument("--role", "-r", help="KERI enpoint authorization role.",
                    required=True)
parser.add_argument("--eid", "-e", help="qualified base64 of AID to authorize with new role for the AID identified "
                                        "by alias",
                    required=True)
parser.add_argument("--time", help="timestamp for the end auth", required=False, default=None)


def add_end(args):
    """ Command line tool for adding endpoint role authorizations

    """
    ld = RoleDoer(name=args.name,
                  base=args.base,
                  alias=args.alias,
                  bran=args.bran,
                  role=args.role,
                  eid=args.eid,
                  timestamp=args.time)
    return [ld]


class RoleDoer(doing.DoDoer):

    def __init__(self, name, base, alias, bran, role, eid, timestamp=None):
        self.role = role
        self.eid = eid
        self.timestamp = timestamp

        self.hby = existing.setupHby(name=name, base=base, bran=bran)
        self.hab = self.hby.habByName(alias)
        self.witpub = WitnessPublisher(hby=self.hby)
        self.postman = forwarding.Poster(hby=self.hby)
        notifier = Notifier(self.hby)
        mux = grouping.Multiplexor(self.hby, notifier=notifier)
        exc = exchanging.Exchanger(hby=self.hby, handlers=[])
        grouping.loadHandlers(exc, mux)

        mbx = indirecting.MailboxDirector(hby=self.hby, topics=["/receipt", "/multisig", "/replay"], exc=exc)

        if self.hab is None:
            raise kering.ConfigurationError(f"unknown alias={alias}")

        self.toRemove = [self.witpub, self.postman, mbx]

        super(RoleDoer, self).__init__(doers=self.toRemove + [doing.doify(self.roleDo)])

    def roleDo(self, tymth, tock=0.0):
        """ Export any end reply messages previous saved for the provided AID

        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock (float): injected initial tock value

        Returns:  doifiable Doist compatible generator method

        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        data = dict(cid=self.hab.pre, role=self.role, eid=self.eid)

        route = "/end/role/add"
        msg = self.hab.reply(route=route, data=data, stamp=self.timestamp)

        parsing.Parser().parse(ims=bytes(msg), kvy=self.hab.kvy, rvy=self.hab.rvy)

        if isinstance(self.hab, habbing.GroupHab):
            smids = self.hab.db.signingMembers(pre=self.hab.pre)
            smids.remove(self.hab.mhab.pre)

            for recp in smids:  # this goes to other participants only as a signaling mechanism
                exn, atc = grouping.multisigRpyExn(ghab=self.hab, rpy=msg)
                self.postman.send(src=self.hab.mhab.pre,
                                  dest=recp,
                                  topic="multisig",
                                  serder=exn,
                                  attachment=atc)

        while not self.hab.loadEndRole(cid=self.hab.pre, role=self.role, eid=self.eid):
            yield self.tock

        self.witpub.msgs.append(dict(pre=self.hab.pre, msg=bytes(msg)))

        while not self.witpub.cues:
            yield self.tock

        print(f"End role authorization added for role {self.role}")

        self.remove(self.toRemove)
        return



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""
import argparse

from hio import help
from hio.base import doing

from keri.app import connecting
from keri.app.cli.common import existing
from keri.kering import ConfigurationError, Roles

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='List current mailboxes')
parser.set_defaults(handler=lambda args: handle(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--alias', '-a', help='human readable alias for the identifier to whom the credential was issued',
                    required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran


def handle(args):
    """ Command line handler for adding an aid to a watcher's list of AIds to watch

    Parameters:
        args(Namespace): parsed command line arguments

    """

    kwa = dict(args=args)
    return [doing.doify(listMailboxes, **kwa)]


def listMailboxes(tymth, tock=0.0, **opts):
    """ Command line status handler

    """
    _ = (yield tock)
    args = opts["args"]
    name = args.name
    alias = args.alias
    base = args.base
    bran = args.bran

    try:
        with existing.existingHby(name=name, base=base, bran=bran) as hby:
            org = connecting.Organizer(hby=hby)
            if alias is None:
                alias = existing.aliasInput(hby)

            hab = hby.habByName(alias)

            for (aid, role, eid), ender in hab.db.ends.getItemIter(keys=(hab.pre, Roles.mailbox)):
                if ender.allowed:
                    contact = org.get(eid)
                    print(f"{contact['alias']}: {eid}")

    except ConfigurationError as e:
        print(f"identifier prefix for {name} does not exist, incept must be run first", )
        return -1



# -*- coding: utf-8 -*-
"""
keri.core.counting module

Provides versioning support for Counter classes and codes
"""
import copy

from dataclasses import dataclass, astuple, asdict
from collections import namedtuple

from ..help import helping
from ..help.helping import sceil
from ..help.helping import (intToB64,  b64ToInt, codeB64ToB2, codeB2ToB64, Reb64,
                            nabSextets)

from .. import kering
from ..kering import (Versionage, Vrsn_1_0, Vrsn_2_0)

from ..core.coring import MapDom



@dataclass(frozen=True)
class GenusCodex:
    """GenusCodex is codex of protocol genera for code table.

    Only provide defined codes.
    Undefined are left out so that inclusion(exclusion) via 'in' operator works.
    """
    KERI_ACDC_SPAC: str = '--AAA'  # KERI, ACDC, and  SPAC Protocol Stacks share the same tables
    KERI: str = '--AAA'  # KERI and ACDC Protocol Stacks share the same tables
    ACDC: str = '--AAA'  # KERI and ACDC Protocol Stacks share the same tables
    SPAC: str = '--AAA'  # KERI and ACDC Protocol Stacks share the same tables


    def __iter__(self):
        return iter(astuple(self))  # enables inclusion test with "in"
        # duplicate values above just result in multiple entries in tuple so
        # in inclusion still works

GenDex = GenusCodex()  # Make instance


@dataclass(frozen=True)
class CounterCodex_1_0(MapDom):
    """
    CounterCodex is codex hard (stable) part of all counter derivation codes.
    Only provide defined codes.
    Undefined are left out so that inclusion(exclusion) via 'in' operator works.

    As subclass of MapCodex can get codes with item syntax using tag variables.
    Example: codex[tag]
    """
    ControllerIdxSigs: str = '-A'  # Qualified Base64 Indexed Signature.
    WitnessIdxSigs: str = '-B'  # Qualified Base64 Indexed Signature.
    NonTransReceiptCouples: str = '-C'  # Composed Base64 Couple, pre+cig.
    TransReceiptQuadruples: str = '-D'  # Composed Base64 Quadruple, pre+snu+dig+sig.
    FirstSeenReplayCouples: str = '-E'  # Composed Base64 Couple, fnu+dts.
    TransIdxSigGroups: str = '-F'  # Composed Base64 Group, pre+snu+dig+ControllerIdxSigs group.
    SealSourceCouples: str = '-G'  # Composed Base64 couple, snu+dig of given delegator/issuer/transaction event
    TransLastIdxSigGroups: str = '-H'  # Composed Base64 Group, pre+ControllerIdxSigs group.
    SealSourceTriples: str = '-I'  # Composed Base64 triple, pre+snu+dig of anchoring source event
    SadPathSigGroups: str = '-J'  # Composed Base64 Group path+TransIdxSigGroup of SAID of content
    RootSadPathSigGroups: str = '-K'  # Composed Base64 Group, root(path)+SaidPathCouples
    PathedMaterialGroup: str = '-L'  # Composed Grouped Pathed Material Quadlet (4 char each)
    BigPathedMaterialGroup: str = '-0L'  # Composed Grouped Pathed Material Quadlet (4 char each)
    AttachmentGroup: str = '-V'  # Composed Grouped Attached Material Quadlet (4 char each)
    BigAttachmentGroup: str = '-0V'  # Composed Grouped Attached Material Quadlet (4 char each)
    ESSRPayloadGroup: str = '-Z'  # ESSR Payload Group, dig of content+Texter group
    KERIACDCGenusVersion: str = '--AAA'  # KERI ACDC Protocol Stack CESR Version


    def __iter__(self):
        return iter(astuple(self))  # enables value not key inclusion test with "in"

CtrDex_1_0 = CounterCodex_1_0()

@dataclass(frozen=True)
class CounterCodex_2_0(MapDom):
    """
    CounterCodex is codex hard (stable) part of all counter derivation codes.
    Only provide defined codes.
    Undefined are left out so that inclusion(exclusion) via 'in' operator works.

    As subclass of MapCodex can get codes with item syntax using tag variables.
    Example: codex[tag]
    """
    GenericGroup: str = '-A'  # Generic Group (Universal with Override).
    BigGenericGroup: str = '-0A'  # Big Generic Group (Universal with Override).
    MessageGroup: str = '-B'  # Message Body plus Attachments Group (Universal with Override).
    BigMessageGroup: str = '-0B'  # Big Message Body plus Attachments Group (Universal with Override).
    AttachmentGroup: str = '-C'  # Message Attachments Only Group (Universal with Override).
    BigAttachmentGroup: str = '-0C'  # Big Attachments Only Group (Universal with Override).
    DatagramSegmentGroup: str = '-D'  # Datagram Segment Group (Universal).
    BigDatagramSegmentGroup: str = '-0D'  # Big Datagram Segment Group (Universal).
    ESSRWrapperGroup: str = '-E'  # ESSR Wrapper Group (Universal).
    BigESSRWrapperGroup: str = '-0E'  # Big ESSR Wrapper Group (Universal).
    FixedMessageBodyGroup: str = '-F'  # Fixed Field Message Body Group (Universal).
    BigFixedMessageBodyGroup: str = '-0F'  # Big Fixed Field Message Body Group (Universal).
    MapMessageBodyGroup: str = '-G'  # Field Map Message Body Group (Universal).
    BigMapMessageBodyGroup: str = '-0G'  # Big Field Map Message Body Group (Universal).
    GenericMapGroup: str = '-H'  # Generic Field Map Group (Universal).
    BigGenericMapGroup: str = '-0H'  # Big Generic Field Map Group (Universal).
    GenericListGroup: str = '-I'  # Generic List Group (Universal).
    BigGenericListGroup: str = '-0I'  # Big Generic List Group (Universal).
    ControllerIdxSigs: str = '-J'  # Controller Indexed Signature(s) of qb64.
    BigControllerIdxSigs: str = '-0J'  # Big Controller Indexed Signature(s) of qb64.
    WitnessIdxSigs: str = '-K'  # Witness Indexed Signature(s) of qb64.
    BigWitnessIdxSigs: str = '-0K'  # Big Witness Indexed Signature(s) of qb64.
    NonTransReceiptCouples: str = '-L'  # NonTrans Receipt Couple(s), pre+cig.
    BigNonTransReceiptCouples: str = '-0L'  # Big NonTrans Receipt Couple(s), pre+cig.
    TransReceiptQuadruples: str = '-M'  # Trans Receipt Quadruple(s), pre+snu+dig+sig.
    BigTransReceiptQuadruples: str = '-0M'  # Big Trans Receipt Quadruple(s), pre+snu+dig+sig.
    FirstSeenReplayCouples: str = '-N'  # First Seen Replay Couple(s), fnu+dts.
    BigFirstSeenReplayCouples: str = '-0N'  # First Seen Replay Couple(s), fnu+dts.
    TransIdxSigGroups: str = '-O'  # Trans Indexed Signature Group(s), pre+snu+dig+CtrControllerIdxSigs of qb64.
    BigTransIdxSigGroups: str = '-0O'  # Big Trans Indexed Signature Group(s), pre+snu+dig+CtrControllerIdxSigs of qb64.
    TransLastIdxSigGroups: str = '-P'  # Trans Last Est Evt Indexed Signature Group(s), pre+CtrControllerIdxSigs of qb64.
    BigTransLastIdxSigGroups: str = '-0P'  # Big Trans Last Est Evt Indexed Signature Group(s), pre+CtrControllerIdxSigs of qb64.
    SealSourceCouples: str = '-Q'  # Seal Source Couple(s), snu+dig of source sealing or sealed event.
    BigSealSourceCouples: str = '-0Q'  # Seal Source Couple(s), snu+dig of source sealing or sealed event.
    SealSourceTriples: str = '-R'  # Seal Source Triple(s), pre+snu+dig of source sealing or sealed event.
    BigSealSourceTriples: str = '-0R'  # Seal Source Triple(s), pre+snu+dig of source sealing or sealed event.
    PathedMaterialGroup: str = '-S'  # Pathed Material Group.
    BigPathedMaterialGroup: str = '-0S'  # Big Pathed Material Group.
    SadPathSigGroups: str = '-T'  # SAD Path Group(s) sadpath+CtrTransIdxSigGroup(s) of SAID qb64 of content.
    BigSadPathSigGroups: str = '-0T'  # Big SAD Path Group(s) sadpath+CtrTransIdxSigGroup(s) of SAID qb64 of content.
    RootSadPathSigGroups: str = '-U'  # Root Path SAD Path Group(s), rootpath+SadPathGroup(s).
    BigRootSadPathSigGroups: str = '-0U'  # Big Root Path SAD Path Group(s), rootpath+SadPathGroup(s).
    DigestSealSingles: str = '-V'  # Digest Seal Single(s), dig of sealed data.
    BigDigestSealSingles: str = '-0V'  # Big Digest Seal Single(s), dig of sealed data.
    MerkleRootSealSingles: str = '-W'  # Merkle Tree Root Digest Seal Single(s), dig of sealed data.
    BigMerkleRootSealSingles: str = '-0W'  # Merkle Tree Root Digest Seal Single(s), dig of sealed data.
    BackerRegistrarSealCouples: str = '-X'  # Backer Registrar Seal Couple(s), brid+dig of sealed data.
    BigBackerRegistrarSealCouples: str = '-0X'  # Big Backer Registrar Seal Couple(s), brid+dig of sealed data.
    SealSourceLastSingles: str = '-Y'  # Seal Source Couple(s), pre of last source sealing or sealed event.
    BigSealSourceLastSingles: str = '-0Y'  # Big Seal Source Couple(s), pre of last source sealing or sealed event.
    ESSRPayloadGroup: str = '-Z'  # ESSR Payload Group.
    BigESSRPayloadGroup: str = '-0Z'  # Big ESSR Payload Group.
    KERIACDCGenusVersion: str = '--AAA'  # KERI ACDC Stack CESR Protocol Genus Version (Universal)

    def __iter__(self):
        return iter(astuple(self))  # enables value not key inclusion test with "in"

CtrDex_2_0 = CounterCodex_2_0()

# CodeNames  is tuple of codes names given by attributes of union of codices
CodeNames = tuple(asdict(CtrDex_2_0) | asdict(CtrDex_1_0))
# Codens  is namedtuple of CodeNames where its names are the code names
# Codens enables using the attributes of the named tuple to specify a code by
# name (indirection) so that changes in the code itself do not break the
# creation of a counter. Enables specifying a counter by the code name not the
# code itself. The code may change between versions but the code name does not.
Codenage = namedtuple("Codenage", CodeNames, defaults=CodeNames)
Codens = Codenage()


@dataclass(frozen=True)
class SealCodex_2_0(MapDom):
    """
    SealCodex_2_0 is codex of seal counter derivation codes.
    Only provide defined codes.
    Undefined are left out so that inclusion(exclusion) via 'in' operator works.

    As subclass of MapCodex can get codes with item syntax using tag variables.
    Example: codex[tag]
    """
    SealSourceCouples: str = '-Q'  # Seal Source Couple(s), snu+dig of source sealing or sealed event.
    BigSealSourceCouples: str = '-0Q'  # Seal Source Couple(s), snu+dig of source sealing or sealed event.
    SealSourceTriples: str = '-R'  # Seal Source Triple(s), pre+snu+dig of source sealing or sealed event.
    BigSealSourceTriples: str = '-0R'  # Seal Source Triple(s), pre+snu+dig of source sealing or sealed event.
    DigestSealSingles: str = '-V'  # Digest Seal Single(s), dig of sealed data.
    BigDigestSealSingles: str = '-0V'  # Big Digest Seal Single(s), dig of sealed data.
    MerkleRootSealSingles: str = '-W'  # Merkle Tree Root Digest Seal Single(s), dig of sealed data.
    BigMerkleRootSealSingles: str = '-0W'  # Merkle Tree Root Digest Seal Single(s), dig of sealed data.
    BackerRegistrarSealCouples: str = '-X'  # Backer Registrar Seal Couple(s), brid+dig of sealed data.
    BigBackerRegistrarSealCouples: str = '-0X'  # Big Backer Registrar Seal Couple(s), brid+dig of sealed data.
    SealSourceLastSingles: str = '-Y'  # Seal Source Couple(s), pre of last source sealing event.
    BigSealSourceLastSingles: str = '-0Y'  # Big Seal Source Couple(s), pre of last source sealing event.

    def __iter__(self):
        return iter(astuple(self))  # enables value not key inclusion test with "in"

SealDex_2_0 = SealCodex_2_0()

# namedtuple for size entries in Counter derivation code tables
# hs is the hard size int number of chars in hard (stable) part of code
# ss is the soft size int number of chars in soft (unstable) part of code
# fs is the full size int number of chars in code
Cizage = namedtuple("Cizage", "hs ss fs")


class Counter:
    """
    Counter is fully qualified cryptographic material primitive base class for
    counter primitives (framing composition grouping count codes).

    Sub classes are derivation code and key event element context specific.

    Includes the following attributes and properties:

    Class Attributes:
        Codes (dict): nested of codexes keyed by major and minor version
        Names (dict): nested of map of code names to codes keyed by
                        major and minor version
        Hards (dict): of hard code sizes keyed by text domain selector
        Bards (dict): of hard code sizes keyed by binary domain selector
        Sizes (dict): of size tables keyed by version. Size table is dict
                      of Sizages keyed by hard code

    Attributes:


    Properties:
        .version (Versionage): current CESR code table protocol genus version
        .codes (CounterCodex_1_0 | CounterCodex_1_0): version specific codex
        .sizes (dict): version specific sizes table
        .code (str): hard part of derivation code to indicate cypher suite
        .raw (bytes): crypto material only without code
        .pad  (int): number of pad chars given raw
        .count (int): count of quadlets/triplets of following framed material
                      (not including code)
        .qb64 (str | bytes | bytearray): in Base64 fully qualified with
                                          derivation code + crypto mat
        .qb64b (bytes | bytearray): in Base64 fully qualified with
                                    derivation code + crypto mat
        .qb2  (bytes | bytearray): in binary with derivation code +
                                  crypto material

    Hidden:
        ._version (Versionage): value for .version property
        ._codes (CounterCodex_1_0 | CounterCodex_1_0): version specific codex
        ._sizes (dict): version specific sizes table
        ._code (str): value for .code property
        ._raw (bytes): value for .raw property
        ._count (int): value for .count property


    Versioning:
        CESR Genus specific code tables have a major and a minor version.

        For a given major version all minor versions must be backwards compatible.
        This means that minor version changes to tables are append only. New
        codes may be added but no existing codes may be changed. This means that
        a given implementation need only use use the latest minor version of
        the code table for a given major version when generating or parsing a
        primitive or group. Assuming the major versions match, when parsing,
        a primitive, when that primitive was generated with a later minor version
        than the implementation supports then it will not be recognized and
        raise an error. But if a primitive was generated with any earlier minor
        version than the version the implementation supports then the primitive
        will parse correctly using any later minor version of the code table.

        Likewise a given protocol stack may have message bodies that carry
        a major and a minor version.

        A given CESR Genus and a given Protocol message stack may be paired in
        order to synchronize versioning between the two when the message bodies
        use primitives and or groups defined by codes in the CESR Genus table.

        In this case pairing is between the CESR Genus labeled KERI_ACDC_SPAC
        and the message body protocol stack labeled KERI/ACDC/SPAC

        The two versions, CESR Genus and Protocol Stack, may be synchronized in
        the following way:

        * Major versions must match or be compatible

        * Minor versions may differ but must be compatible within a
        major version.

        Importantly the CESR code table version may not be included in the
        message body itself but only provided in the surrounding CESR stream.
        This means the code table version used by a message body may not be
        signed. Therefore the receiver of a message body with embedded CESR
        primitives and groups must be protected from a CESR code table genus
        version malleability attack.

        When the major versions of the CESR code table and protocol stack
        match, the signed embedded protocol stack major version protects
        the receiver from a major version malleability attack on the CESR
        code table. Otherwise the major versions must be compatible in a way
        that does not allow malleability. For example the set of allowed codes
        for a given message protocol version are compatible across CESR code
        table major versions.

        This, however, does not protect the receiver of a message body from
        a minor version malleability attack on the CESR code table.
        Nevertheless, the requirement that all minor versions of a CESR code
        table for a given major version must be backwards compatible,
        does indeed provide this protection.

        Either, the receiver of the message body recognizes exactly
        all primitives and groups in the message body because the CESR code
        table minor version supported by the receiver is greater than or equal
        to that used by the the minor version of the sender or any unsupported
        (later appended) primitives or group codes will be unrecognized by
        the received thereby raising an error that results in the message being
        dropped.

    """
    Codes = \
    {
        Vrsn_1_0.major: \
        {
            Vrsn_1_0.minor: CtrDex_1_0,
        },
        Vrsn_2_0.major: \
        {
            Vrsn_2_0.minor: CtrDex_2_0,
        },
    }


    # invert dataclass codenames: codes to dict codes: codenames
    Names = copy.deepcopy(Codes)  # make deep nested copy so can invert nested values
    for minor in Names.values():
        for key in minor:
            minor[key] = {val: key for key, val in asdict(minor[key]).items()}




    # Hards table maps from bytes Base64 first two code chars to int of
    # hard size, hs,(stable) of code. The soft size, ss, (unstable) for Counter
    # is always > 0 and hs + ss = fs always
    Hards = ({('-' + chr(c)): 2 for c in range(65, 65 + 26)})
    Hards.update({('-' + chr(c)): 2 for c in range(97, 97 + 26)})
    Hards.update([('-0', 3)])
    Hards.update([('--', 5)])

    # Bards table maps to hard size, hs, of code from bytes holding sextets
    # converted from first two code char. Used for ._bexfil.
    Bards = ({codeB64ToB2(c): hs for c, hs in Hards.items()})

    # Sizes table indexes size tables first by major version and then by
    # lastest minor version
    # Each size table maps hs chars of code to Cizage namedtuple of (hs, ss, fs)
    # where hs is hard size, ss is soft size, and fs is full size
    # soft size, ss, should always be  > 0 and hs+ss=fs for Counter
    Sizes = \
    {
        Vrsn_1_0.major: \
        {
            Vrsn_1_0.minor: \
            {
                '-A': Cizage(hs=2, ss=2, fs=4),
                '-B': Cizage(hs=2, ss=2, fs=4),
                '-C': Cizage(hs=2, ss=2, fs=4),
                '-D': Cizage(hs=2, ss=2, fs=4),
                '-E': Cizage(hs=2, ss=2, fs=4),
                '-F': Cizage(hs=2, ss=2, fs=4),
                '-G': Cizage(hs=2, ss=2, fs=4),
                '-H': Cizage(hs=2, ss=2, fs=4),
                '-I': Cizage(hs=2, ss=2, fs=4),
                '-J': Cizage(hs=2, ss=2, fs=4),
                '-K': Cizage(hs=2, ss=2, fs=4),
                '-L': Cizage(hs=2, ss=2, fs=4),
                '-0L': Cizage(hs=3, ss=5, fs=8),
                '-V': Cizage(hs=2, ss=2, fs=4),
                '-0V': Cizage(hs=3, ss=5, fs=8),
                '-Z': Cizage(hs=2, ss=2, fs=4),
                '--AAA': Cizage(hs=5, ss=3, fs=8),
            },
        },
        Vrsn_2_0.major: \
        {
            Vrsn_2_0.minor: \
            {
                '-A': Cizage(hs=2, ss=2, fs=4),
                '-0A': Cizage(hs=3, ss=5, fs=8),
                '-B': Cizage(hs=2, ss=2, fs=4),
                '-0B': Cizage(hs=3, ss=5, fs=8),
                '-C': Cizage(hs=2, ss=2, fs=4),
                '-0C': Cizage(hs=3, ss=5, fs=8),
                '-D': Cizage(hs=2, ss=2, fs=4),
                '-0D': Cizage(hs=3, ss=5, fs=8),
                '-E': Cizage(hs=2, ss=2, fs=4),
                '-0E': Cizage(hs=3, ss=5, fs=8),
                '-F': Cizage(hs=2, ss=2, fs=4),
                '-0F': Cizage(hs=3, ss=5, fs=8),
                '-G': Cizage(hs=2, ss=2, fs=4),
                '-0G': Cizage(hs=3, ss=5, fs=8,),
                '-H': Cizage(hs=2, ss=2, fs=4),
                '-0H': Cizage(hs=3, ss=5, fs=8),
                '-I': Cizage(hs=2, ss=2, fs=4),
                '-0I': Cizage(hs=3, ss=5, fs=8),
                '-J': Cizage(hs=2, ss=2, fs=4,),
                '-0J': Cizage(hs=3, ss=5, fs=8),
                '-K': Cizage(hs=2, ss=2, fs=4),
                '-0K': Cizage(hs=3, ss=5, fs=8),
                '-L': Cizage(hs=2, ss=2, fs=4),
                '-0L': Cizage(hs=3, ss=5, fs=8),
                '-M': Cizage(hs=2, ss=2, fs=4),
                '-0M': Cizage(hs=3, ss=5, fs=8),
                '-N': Cizage(hs=2, ss=2, fs=4),
                '-0N': Cizage(hs=3, ss=5, fs=8),
                '-O': Cizage(hs=2, ss=2, fs=4),
                '-0O': Cizage(hs=3, ss=5, fs=8),
                '-P': Cizage(hs=2, ss=2, fs=4),
                '-0P': Cizage(hs=3, ss=5, fs=8),
                '-Q': Cizage(hs=2, ss=2, fs=4),
                '-0Q': Cizage(hs=3, ss=5, fs=8),
                '-R': Cizage(hs=2, ss=2, fs=4),
                '-0R': Cizage(hs=3, ss=5, fs=8),
                '-S': Cizage(hs=2, ss=2, fs=4),
                '-0S': Cizage(hs=3, ss=5, fs=8),
                '-T': Cizage(hs=2, ss=2, fs=4),
                '-0T': Cizage(hs=3, ss=5, fs=8),
                '-U': Cizage(hs=2, ss=2, fs=4),
                '-0U': Cizage(hs=3, ss=5, fs=8),
                '-V': Cizage(hs=2, ss=2, fs=4),
                '-0V': Cizage(hs=3, ss=5, fs=8),
                '-W': Cizage(hs=2, ss=2, fs=4),
                '-0W': Cizage(hs=3, ss=5, fs=8),
                '-X': Cizage(hs=2, ss=2, fs=4),
                '-0X': Cizage(hs=3, ss=5, fs=8),
                '-Y': Cizage(hs=2, ss=2, fs=4),
                '-0Y': Cizage(hs=3, ss=5, fs=8),
                '-Z': Cizage(hs=2, ss=2, fs=4),
                '-0Z': Cizage(hs=3, ss=5, fs=8),
                '--AAA': Cizage(hs=5, ss=3, fs=8),
            },
        },
    }


    def __init__(self, code=None, *, count=None, countB64=None,
                 qb64b=None, qb64=None, qb2=None, strip=False,
                 gvrsn=Vrsn_2_0, **kwa):
        """
        Validate as fully qualified
        Parameters:
            code (str | None):  either stable (hard) part of derivation code or
                                code name. When code name then look up code from
                                ._codes. This allows versioning to change code
                                but keep stable code name.

            count (int | None): count of framed material in quadlets/triplets
                               for composition. Count does not include code.
                               When both count and countB64 are None then count
                               defaults to 1
            countB64 (str | None): count of framed material in quadlets/triplets
                                for composition as Base64 representation of int.
            qb64b (bytes | bytearray | None): fully qualified crypto material text domain
                if code nor tag is provided
            qb64 (str | None) fully qualified crypto material text domain
                if code nor tag not qb64b is provided
            qb2 (bytes | bytearray | None)  fully qualified crypto material binary domain
                if code nor tag not qb64b nor qb54 is provided
            strip (bool):  True means strip counter contents from input stream
                bytearray after parsing qb64b or qb2. False means do not strip.
                default False
            gvrsn (Versionage): instance of genera version of CESR code tables


        Needs either code or qb64b or qb64 or qb2
        Otherwise raises EmptyMaterialError
        When code and count provided then validate that code and count are correct
        Else when qb64b or qb64 or qb2 provided extract and assign
        .code and .count

        """
        if gvrsn.major not in self.Sizes:
            raise kering.InvalidVersionError(f"Unsupported major version="
                                             f"{gvrsn.major}.")

        latest = list(self.Sizes[gvrsn.major])[0]  # get latest minor version
        if gvrsn.minor > latest:
            raise kering.InvalidVersionError(f"Minor version={gvrsn.minor} "
                                             f" exceeds latest supported minor"
                                             f" version={latest}.")

        self._codes = self.Codes[gvrsn.major][latest]  # use latest supported version codes
        self._sizes = self.Sizes[gvrsn.major][latest]  # use latest supported version sizes
        self._version = gvrsn  # provided version may be earlier than supported version


        if code:  # code (hard) provided
             # assumes ._sizes ._codes coherent
            if code not in self._sizes or len(code) < 2:
                try:
                    code = self._codes[code]  # code is code name so look up code
                    if code not in self._sizes or len(code) < 2:
                        raise kering.InvalidCodeError(f"Unsupported {code=}.")
                except Exception as ex:
                    raise kering.InvalidCodeError(f"Unsupported {code=}.") from ex

            hs, ss, fs = self._sizes[code]  # get sizes for code
            cs = hs + ss  # both hard + soft code size
            if hs < 2 or fs != cs or cs % 4:  # fs must be bs and multiple of 4 for count codes
                raise kering.InvalidCodeSizeError(f"Whole code size not full "
                                                  f"size or not multiple of 4. "
                                                  f"{cs=} {fs=}.")

            if count is None:
                count = 1 if countB64 is None else b64ToInt(countB64)

            if code[1] not in ("123456789-_"):  # small [A-Z,a-z] or large [0]
                if ss not in (2, 5):  # not valid dynamic soft sizes
                    raise kering.InvalidVarIndexError(f"Invalid {ss=} "
                                                      f"for {code=}.")
                # dynamically promote code based on count
                if code[1] != '0' and count > (64 ** 2 - 1):  # small code but large count
                    # elevate code due to large count
                    code = f"-0{code[1]}"  # promote hard
                    ss = 5

            if count < 0 or count > (64 ** ss - 1):
                raise kering.InvalidVarIndexError(f"Invalid {count=} for "
                                                  f"{code=} with {ss=}.")

            self._code = code
            self._count = count

        elif qb64b is not None:
            self._exfil(qb64b)
            if strip:  # assumes bytearray
                del qb64b[:self._sizes[self.code].fs]

        elif qb64 is not None:
            self._exfil(qb64)

        elif qb2 is not None:  # rewrite to use direct binary exfiltration
            self._bexfil(qb2)
            if strip:  # assumes bytearray
                del qb2[:self._sizes[self.code].fs * 3 // 4]

        else:
            raise kering.EmptyMaterialError("Improper initialization need either "
                                     "(code and count) or qb64b or "
                                     "qb64 or qb2.")

        self._name = self.Names[gvrsn.major][latest][self.code]

    @property
    def version(self):
        """
        Returns ._version
        Makes .version read only
        """
        return self._version

    @property
    def gvrsn(self):
        """
        Returns .version alias for .version

        """
        return self.version

    @property
    def codes(self):
        """
        Returns ._codes
        Makes .codes read only
        """
        return self._codes


    @property
    def sizes(self):
        """
        Returns ._sizes
        Makes .sizes read only
        """
        return self._sizes

    @property
    def code(self):
        """
        Returns:
            code (str): hard part only of full text code.
                Getter for ._code. Makes .code read only

        Soft part is count
        """
        return self._code

    @property
    def name(self):
        """
        Returns:
            name (str): code name for self.code. Match interface
            for annotation for primitives like Matter

        Getter for ._name. Makes .name read only

        """
        return self._name


    @property
    def hard(self):
        """
        Returns:
            hard (str): hard part only of full text code. Alias for .code.

        """
        return self.code


    @property
    def count(self):
        """
        Returns:
            count (int):  count value in quadlets/triples chars/bytes of material
                framed by counter.
                Getter for ._count. Makes ._count read only
        """
        return self._count


    @property
    def soft(self):
        """
       Returns:
            soft (str):  Base64 soft part of full counter code. Count value in
                quadlets/triples chars/bytes of material framed by counter.
                Converts .count to b64
        """
        _, ss, _ = self.sizes[self.code]
        return intToB64(self._count, l=ss)


    @property
    def both(self):
        """
        Returns:
            both (str):  hard + soft parts of full text code
        """
        return f"{self.hard}{self.soft}"


    @property
    def fullSize(self):
        """
        Returns full size of counter in bytes

        """
        _, _, fs = self.sizes[self.code]  # get from sizes table

        return fs


    @property
    def qb64b(self):
        """
        Property qb64b:
        Returns Fully Qualified Base64 Version encoded as bytes
        Assumes self.raw and self.code are correctly populated
        """
        return self._infil()


    @property
    def qb64(self):
        """
        Property qb64:
        Returns Fully Qualified Base64 Version
        Assumes self.raw and self.code are correctly populated
        """
        return self.qb64b.decode("utf-8")


    @property
    def qb2(self):
        """
        Property qb2:
        Returns Fully Qualified Binary Version Bytes
        """
        return self._binfil()


    def countToB64(self, l=None):
        """ Returns count as Base64 left padded with "A"s
            Parameters:
                l (int | None): minimum number characters including left padding
                    When not provided use the softsize of .code

        """
        if l is None:
            _, ss, _ = self._sizes[self.code]
            l = ss
        return (intToB64(self.count, l=l))


    @staticmethod
    def verToB64(version=None, *, text="", major=0, minor=0):
        """ Converts version to Base64 representation of countB64
        suitable for CESR protocol genus and version

        Returns:
            countB64 (str): suitable for input to Counter

        Example:
            Counter(countB64=Counter.verToB64(verstr = "1.0"))

        Parameters:
            version (Versionage): instange of namedtuple
                         Versionage(major=major,minor=minor)
            text (str): text format of version as dotted decimal "major.minor"
            major (int): When version is None and verstr is empty then use major minor
                        range [0, 63] for one Base64 character
            minor (int): When version is None and verstr is  empty then use major minor
                        range [0, 4095] for two Base64 characters

        """
        if version:
            major = version.major
            minor = version.minor

        elif text:
            splits = text.split(".", maxsplit=2)
            splits = [(int(s) if s else 0) for s in splits]
            parts = [major, minor]
            for i in range(2-len(splits),0, -1):  # append missing minor and/or major
                splits.append(parts[-i])
            major = splits[0]
            minor = splits[1]

        if major < 0 or major > 63 or minor < 0 or minor > 4095:
                raise ValueError(f"Out of bounds version = {major}.{minor}.")

        return (f"{intToB64(major)}{intToB64(minor, l=2)}")


    @staticmethod
    def b64ToVer(b64, *, texted=False):
        """ Converts Base64 representation of version to Versionage or
        text dotted decimal format

        default is Versionage

        Returns:
            version (Versionage | str):

        Example:
            Counter(version=Counter.b64ToVer("BAA"))

        Parameters:
            b64 (str): base64 string of three characters Mmm for Major minor
            texted (bool): return text format dotted decimal string


        """
        if not Reb64.match(b64.encode("utf-8")):
            raise ValueError("Invalid Base64.")

        if texted:
            return ".".join([f"{b64ToInt(b64[0])}", f"{b64ToInt(b64[1:3])}"])

        return Versionage(major=b64ToInt(b64[0]), minor=b64ToInt(b64[1:3]))


    def _infil(self):
        """
        Returns fully qualified attached sig base64 bytes computed from
        self.code and self.count.
        """
        code = self.code  # codex value chars hard code
        count = self.count  # index value int used for soft

        hs, ss, fs = self._sizes[code]
        # assumes fs = hs + ss  # both hard + soft size
        # assumes unit tests ensure ._sizes table entries are consistent
        # hs >= 2, ss > 0 fs == hs + ss, not (fs % 4)

        if count < 0 or count > (64 ** ss - 1):
            raise kering.InvalidVarIndexError("Invalid count={} for code={}.".format(count, code))

        # both is hard code + converted count
        both = "{}{}".format(code, intToB64(count, l=ss))

        # check valid pad size for whole code size
        if len(both) % 4:  # no pad
            raise kering.InvalidCodeSizeError("Invalid size = {} of {} not a multiple of 4."
                                       .format(len(both), both))
        # prepending full derivation code with index and strip off trailing pad characters
        return (both.encode("utf-8"))


    def _binfil(self):
        """
        Returns bytes of fully qualified base2 bytes, that is .qb2
        self.code converted to Base2 left shifted with pad bits
        equivalent of Base64 decode of .qb64 into .qb2
        """
        code = self.code  # codex chars hard code
        count = self.count  # index value int used for soft

        hs, ss, fs = self._sizes[code]
        # assumes fs = hs + ss
        # assumes unit tests ensure ._sizes table entries are consistent
        # hs >= 2, ss>0 fs ==  hs + ss, not (fs % 4)

        if count < 0 or count > (64 ** ss - 1):
            raise kering.InvalidVarIndexError("Invalid count={} for code={}.".format(count, code))

        # both is hard code + converted count
        both = "{}{}".format(code, intToB64(count, l=ss))
        if len(both) != fs:
            raise kering.InvalidCodeSizeError("Mismatch code size = {} with table = {}."
                                       .format(fs, len(both)))

        return (codeB64ToB2(both))  # convert to b2 left shift if any


    def _exfil(self, qb64b):
        """
        Extracts self.code and self.count from qualified base64 bytes qb64b
        """
        if not qb64b:  # empty need more bytes
            raise kering.ShortageError("Empty material, Need more characters.")

        first = qb64b[:2]  # extract first two char code selector
        if hasattr(first, "decode"):
            first = first.decode("utf-8")
        if first not in self.Hards:
            if first[0] == '_':
                raise kering.UnexpectedOpCodeError("Unexpected op code start"
                                            "while extracing Counter.")
            else:
                raise kering.UnexpectedCodeError("Unsupported code start ={}.".format(first))

        hs = self.Hards[first]  # get hard code size
        if len(qb64b) < hs:  # need more bytes
            raise kering.ShortageError("Need {} more characters.".format(hs - len(qb64b)))

        hard = qb64b[:hs]  # get hard code
        if hasattr(hard, "decode"):
            hard = hard.decode("utf-8")  # decode converts bytearray/bytes to str
        if hard not in self._sizes:  # Sizes needs str not bytes
            raise kering.UnexpectedCodeError("Unsupported code ={}.".format(hard))

        hs, ss, fs = self._sizes[hard]  # assumes hs consistent in both tables
        # assumes fs = hs + ss  # both hard + soft code size
        # assumes that unit tests on Counter and CounterCodex ensure that
        # .Codes and .Sizes are well formed.
        # hs consistent and hs >= 2 and ss > 0 and fs = hs + ss and not fs % 4

        if len(qb64b) < fs:  # need more bytes
            raise kering.ShortageError("Need {} more characters.".format(fs - len(qb64b)))

        count = qb64b[hs:fs]  # extract count chars
        if hasattr(count, "decode"):
            count = count.decode("utf-8")
        count = b64ToInt(count)  # compute int count

        self._code = hard
        self._count = count


    def _bexfil(self, qb2):
        """
        Extracts self.code and self.count from qualified base2 bytes qb2
        """
        if not qb2:  # empty need more bytes
            raise kering.ShortageError("Empty material, Need more bytes.")

        first = nabSextets(qb2, 2)  # extract first two sextets as code selector
        if first not in self.Bards:
            if first[0] == b'\xfc':  # b64ToB2('_')
                raise kering.UnexpectedOpCodeError("Unexpected  op code start"
                                            "while extracing Matter.")
            else:
                raise kering.UnexpectedCodeError("Unsupported code start sextet={}.".format(first))

        hs = self.Bards[first]  # get code hard size equvalent sextets
        bhs = sceil(hs * 3 / 4)  # bhs is min bytes to hold hs sextets
        if len(qb2) < bhs:  # need more bytes
            raise kering.ShortageError("Need {} more bytes.".format(bhs - len(qb2)))

        hard = codeB2ToB64(qb2, hs)  # extract and convert hard part of code
        if hard not in self._sizes:
            raise kering.UnexpectedCodeError("Unsupported code ={}.".format(hard))

        hs, ss, fs = self._sizes[hard]
        # assumes fs = hs + ss  # both hs and ss
        # assumes that unit tests on Counter and CounterCodex ensure that
        # .Codes and .Sizes are well formed.
        # hs consistent and hs >= 2 and ss > 0 and fs = hs + ss and not fs % 4

        bcs = sceil(fs * 3 / 4)  # bcs is min bytes to hold fs sextets
        if len(qb2) < bcs:  # need more bytes
            raise kering.ShortageError("Need {} more bytes.".format(bcs - len(qb2)))

        both = codeB2ToB64(qb2, fs)  # extract and convert both hard and soft part of code
        count = b64ToInt(both[hs:fs])  # get count

        self._code = hard
        self._count = count

